# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import json
import logging
import math
import os
import time
from collections import defaultdict
from typing import (TYPE_CHECKING,
                    Any,
                    Callable,
                    Dict,
                    List,
                    Optional,
                    Set,
                    Tuple,
                    Union)

from toil.batchSystems.abstractBatchSystem import (AbstractBatchSystem,
                                                   AbstractScalableBatchSystem,
                                                   NodeInfo)
from toil.bus import ClusterDesiredSizeMessage, ClusterSizeMessage
from toil.common import Config
from toil.options.common import defaultTargetTime
from toil.job import JobDescription, ServiceJobDescription
from toil.lib.conversions import bytes2human, human2bytes
from toil.lib.retry import old_retry
from toil.lib.threading import ExceptionalThread
from toil.lib.throttle import throttle
from toil.provisioners.abstractProvisioner import AbstractProvisioner, Shape

if TYPE_CHECKING:
    from toil.leader import Leader
    from toil.provisioners.node import Node

logger = logging.getLogger(__name__)

# Properties of GKE's memory overhead algorithm
EVICTION_THRESHOLD = human2bytes('100MiB')
RESERVE_SMALL_LIMIT = human2bytes('1GiB')
RESERVE_SMALL_AMOUNT = human2bytes('255MiB')
RESERVE_BREAKPOINTS: List[Union[int, float]] = [human2bytes('4GiB'), human2bytes('8GiB'), human2bytes('16GiB'), human2bytes('128GiB'), math.inf]
RESERVE_FRACTIONS = [0.25, 0.2, 0.1, 0.06, 0.02]

# Guess of how much disk space on the root volume is used for the OS and essential container images
OS_SIZE = human2bytes('5G')

# Define a type for an explanation of why a job can't fit on a node.
# Consists of a resource name and a constraining value for that resource.
FailedConstraint = Tuple[str, Union[int, float, bool]]

class BinPackedFit:
    """
    If jobShapes is a set of tasks with run requirements (mem/disk/cpu), and nodeShapes is a sorted
    list of available computers to run these jobs on, this function attempts to return a dictionary
    representing the minimum set of computerNode computers needed to run the tasks in jobShapes.

    Uses a first fit decreasing (FFD) bin packing like algorithm to calculate an approximate minimum
    number of nodes that will fit the given list of jobs.  BinPackingFit assumes the ordered list,
    nodeShapes, is ordered for "node preference" outside of BinPackingFit beforehand. So when
    virtually "creating" nodes, the first node within nodeShapes that fits the job is the one
    that's added.

    :param list nodeShapes: The properties of an atomic node allocation, in terms of wall-time,
                            memory, cores, disk, and whether it is preemptible or not.
    :param targetTime: The time before which all jobs should at least be started.

    :returns: The minimum number of minimal node allocations estimated to be required to run all
              the jobs in jobShapes.
    """
    nodeReservations: Dict[Shape, List['NodeReservation']]

    def __init__(self, nodeShapes: List[Shape], targetTime: float = defaultTargetTime) -> None:
        self.nodeShapes = sorted(nodeShapes)
        self.targetTime = targetTime
        self.nodeReservations = {nodeShape: [] for nodeShape in nodeShapes}

    def binPack(self, jobShapes: List[Shape]) -> Dict[Shape, List[FailedConstraint]]:
        """
        Pack a list of jobShapes into the fewest nodes reasonable.
        
        Can be run multiple times.
        
        Returns any distinct Shapes that did not fit, mapping to reasons they did not fit.
        """
        # TODO: Check for redundancy with batchsystems.mesos.JobQueue() sorting
        logger.debug('Running bin packing for node shapes %s and %s job(s).',
                     self.nodeShapes, len(jobShapes))
        # Sort in descending order from largest to smallest. The FFD like-strategy will pack the
        # jobs in order from longest to shortest.
        jobShapes.sort()
        jobShapes.reverse()
        assert len(jobShapes) == 0 or jobShapes[0] >= jobShapes[-1]
        could_not_fit = {}
        for jS in jobShapes:
            rejection = self.addJobShape(jS)
            if rejection is not None:
                # The job bounced because it didn't fit in any bin.
                could_not_fit[rejection[0]] = rejection[1]
        return could_not_fit

    def addJobShape(self, jobShape: Shape) -> Optional[Tuple[Shape, List[FailedConstraint]]]:
        """
        Add the job to the first node reservation in which it will fit. (This
        is the bin-packing aspect).
        
        Returns the job shape again, and a list of failed constraints, if it did not fit.
        """
        chosenNodeShape = None
        for nodeShape in self.nodeShapes:
            if NodeReservation(nodeShape).fits(jobShape):
                # This node shape is the first that fits this jobShape
                chosenNodeShape = nodeShape
                break

        if chosenNodeShape is None:
            logger.debug("Couldn't fit job with requirements %s into any nodes in the nodeTypes "
                         "list.", jobShape)
            # Go back and debug why this happened.
            fewest_constraints: Optional[List[FailedConstraint]] = None
            for shape in self.nodeShapes:
                failures = NodeReservation(nodeShape).get_failed_constraints(jobShape)
                if fewest_constraints is None or len(failures) < len(fewest_constraints):
                    # This was closer to fitting.
                    # TODO: Check the actual constraint values so we don't tell
                    # the user to raise the memory on the smallest machine?
                    fewest_constraints = failures
            
            return jobShape, fewest_constraints if fewest_constraints is not None else []

        # grab current list of job objects appended to this instance type
        nodeReservations = self.nodeReservations[chosenNodeShape]
        for nodeReservation in nodeReservations:
            if nodeReservation.attemptToAddJob(jobShape, chosenNodeShape, self.targetTime):
                # We succeeded adding the job to this node reservation. Now we're done.
                return None

        reservation = NodeReservation(chosenNodeShape)
        currentTimeAllocated = chosenNodeShape.wallTime
        adjustEndingReservationForJob(reservation, jobShape, 0)
        self.nodeReservations[chosenNodeShape].append(reservation)

        # Extend the reservation if necessary to cover the job's entire runtime.
        while currentTimeAllocated < jobShape.wallTime:
            extendThisReservation = NodeReservation(reservation.shape)
            currentTimeAllocated += chosenNodeShape.wallTime
            reservation.nReservation = extendThisReservation
            reservation = extendThisReservation
        return None

    def getRequiredNodes(self) -> Dict[Shape, int]:
        """Return a dict from node shape to number of nodes required to run the packed jobs."""
        return {
            nodeShape: len(self.nodeReservations[nodeShape])
            for nodeShape in self.nodeShapes
        }


class NodeReservation:
    """
    The amount of resources that we expect to be available on a given node at each point in time.

    To represent the resources available in a reservation, we represent a
    reservation as a linked list of NodeReservations, each giving the
    resources free within a single timeslice.
    """

    def __init__(self, shape: Shape) -> None:
        # The wall-time of this slice and resources available in this timeslice
        self.shape = shape
        # The next portion of the reservation (None if this is the end)
        self.nReservation: Optional[NodeReservation] = None

    def __str__(self) -> str:
        return "-------------------\n" \
               "Current Reservation\n" \
               "-------------------\n" \
               "Shape wallTime: %s\n" \
               "Shape memory: %s\n" \
               "Shape cores: %s\n" \
               "Shape disk: %s\n" \
               "Shape preempt: %s\n" \
               "\n" \
               "nReserv wallTime: %s\n" \
               "nReserv memory: %s\n" \
               "nReserv cores: %s\n" \
               "nReserv disk: %s\n" \
               "nReserv preempt: %s\n" \
               "\n" \
               "Time slices: %s\n" \
               "\n" % \
               (self.shape.wallTime,
                self.shape.memory,
                self.shape.cores,
                self.shape.disk,
                self.shape.preemptible,
                self.nReservation.shape.wallTime if self.nReservation is not None else str(None),
                self.nReservation.shape.memory if self.nReservation is not None else str(None),
                self.nReservation.shape.cores if self.nReservation is not None else str(None),
                self.nReservation.shape.disk if self.nReservation is not None else str(None),
                self.nReservation.shape.preemptible if self.nReservation is not None else str(None),
                str(len(self.shapes())))

    def get_failed_constraints(self, job_shape: Shape) -> List[FailedConstraint]:
        """
        Check if a job shape's resource requirements will fit within this allocation.
        
        If the job does *not* fit, returns the failing constraints: the resources
        that can't be accomodated, and the limits that were hit.
        
        If the job *does* fit, returns an empty list.
        
        Must always agree with fits()! This codepath is slower and used for diagnosis.
        """
        
        failures: List[FailedConstraint] = []
        if job_shape.memory > self.shape.memory:
            failures.append(("memory", self.shape.memory))
        if job_shape.cores > self.shape.cores:
            failures.append(("cores", self.shape.cores))
        if job_shape.disk > self.shape.disk:
            failures.append(("disk", self.shape.disk))
        if not job_shape.preemptible and self.shape.preemptible:
            failures.append(("preemptible", self.shape.preemptible))
        return failures
    
    def fits(self, jobShape: Shape) -> bool:
        """Check if a job shape's resource requirements will fit within this allocation."""
        return jobShape.memory <= self.shape.memory and \
               jobShape.cores <= self.shape.cores and \
               jobShape.disk <= self.shape.disk and \
               (jobShape.preemptible or not self.shape.preemptible)

    def shapes(self) -> List[Shape]:
        """Get all time-slice shapes, in order, from this reservation on."""
        shapes = []
        curRes: Optional[NodeReservation] = self
        while curRes is not None:
            shapes.append(curRes.shape)
            curRes = curRes.nReservation
        return shapes

    def subtract(self, jobShape: Shape) -> None:
        """Subtract the resources necessary to run a jobShape from the reservation."""
        self.shape = Shape(self.shape.wallTime,
                           self.shape.memory - jobShape.memory,
                           self.shape.cores - jobShape.cores,
                           self.shape.disk - jobShape.disk,
                           self.shape.preemptible)

    def attemptToAddJob(
        self, jobShape: Shape, nodeShape: Shape, targetTime: float
    ) -> bool:
        """
        Attempt to pack a job into this reservation timeslice and/or the reservations after it.

        jobShape is the Shape of the job requirements, nodeShape is the Shape of the node this
        is a reservation for, and targetTime is the maximum time to wait before starting this job.
        """
        # starting slice of time that we can fit in so far
        startingReservation: Optional[NodeReservation] = self
        # current end of the slices we can fit in so far
        endingReservation = self
        # the amount of runtime of the job currently covered by slices
        availableTime: float = 0
        # total time from when the instance started up to startingReservation
        startingReservationTime: float = 0

        while True:
            # True == can run the job (resources & preemptible only; NO time)
            if endingReservation.fits(jobShape):
                # add the time left available on the reservation
                availableTime += endingReservation.shape.wallTime
                # does the job time fit in the reservation's remaining time?
                if availableTime >= jobShape.wallTime:
                    timeSlice: float = 0
                    while (startingReservation != endingReservation):
                        # removes resources only (NO time) from startingReservation
                        startingReservation.subtract(jobShape)  # type: ignore
                        # set aside the timeSlice
                        timeSlice += startingReservation.shape.wallTime  # type: ignore
                        startingReservation = startingReservation.nReservation  # type: ignore
                    assert jobShape.wallTime - timeSlice <= startingReservation.shape.wallTime
                    adjustEndingReservationForJob(endingReservation, jobShape, timeSlice)
                    # Packed the job.
                    return True

                # If the job would fit, but is longer than the total node allocation
                # extend the node allocation
                elif endingReservation.nReservation == None and startingReservation == self:
                    # Extend the node reservation to accommodate jobShape
                    endingReservation.nReservation = NodeReservation(nodeShape)
            # can't run the job with the current resources
            else:
                if startingReservationTime + availableTime + endingReservation.shape.wallTime <= targetTime:
                    startingReservation = endingReservation.nReservation
                    startingReservationTime += availableTime + endingReservation.shape.wallTime
                    availableTime = 0
                else:
                    break
            nextEndingReservation = endingReservation.nReservation
            if nextEndingReservation is None:
                # Reached the end of the reservation without success so stop trying to
                # add to reservation
                return False
            else:
                endingReservation = nextEndingReservation

        # Couldn't pack the job.
        return False


def adjustEndingReservationForJob(
    reservation: NodeReservation, jobShape: Shape, wallTime: float
) -> None:
    """
    Add a job to an ending reservation that ends at wallTime.

    (splitting the reservation if the job doesn't fill the entire timeslice)
    """
    if jobShape.wallTime - wallTime < reservation.shape.wallTime:
        # This job only partially fills one of the slices. Create a new slice.
        reservation.shape, nS = split(reservation.shape, jobShape, jobShape.wallTime - wallTime)
        nS.nReservation = reservation.nReservation
        reservation.nReservation = nS
    else:
        # This job perfectly fits within the boundaries of the slices.
        reservation.subtract(jobShape)


def split(
    nodeShape: Shape, jobShape: Shape, wallTime: float
) -> Tuple[Shape, NodeReservation]:
    """
    Partition a node allocation into two to fit the job.

    Returning the modified shape of the node and a new node reservation for
    the extra time that the job didn't fill.
    """
    return (Shape(wallTime,
                  nodeShape.memory - jobShape.memory,
                  nodeShape.cores - jobShape.cores,
                  nodeShape.disk - jobShape.disk,
                  nodeShape.preemptible),
            NodeReservation(Shape(nodeShape.wallTime - wallTime,
                                  nodeShape.memory,
                                  nodeShape.cores,
                                  nodeShape.disk,
                                  nodeShape.preemptible)))


def binPacking(nodeShapes: List[Shape], jobShapes: List[Shape], goalTime: float) -> Tuple[Dict[Shape, int], Dict[Shape, List[FailedConstraint]]]:
    """
    Using the given node shape bins, pack the given job shapes into nodes to
    get them done in the given amount of time.
    
    Returns a dict saying how many of each node will be needed, a dict from job
    shapes that could not fit to reasons why.
    """
    bpf = BinPackedFit(nodeShapes, goalTime)
    could_not_fit = bpf.binPack(jobShapes)
    return bpf.getRequiredNodes(), could_not_fit


class ClusterScaler:
    def __init__(
        self, provisioner: AbstractProvisioner, leader: "Leader", config: Config
    ) -> None:
        """
        Class manages automatically scaling the number of worker nodes.

        :param provisioner: Provisioner class with functions to create/remove resources (i.e. AWS).
        :param leader: Leader class which runs the main workflow loop until complete.
        :param config: Config object from which to draw parameters.
        """
        self.provisioner = provisioner
        self.leader = leader
        self.config = config
        self.static: Dict[bool, Dict[str, "Node"]] = {}
        
        # If we encounter a Shape of job that we don't think we can run, call
        # these callbacks with the Shape that didn't fit and the Shapes that
        # were available.
        self.on_too_big: List[Callable[[Shape, List[Shape]], Any]] = []

        # Dictionary of job names to their average runtime, used to estimate wall time of queued
        # jobs for bin-packing
        self.jobNameToAvgRuntime: Dict[str, float] = {}
        self.jobNameToNumCompleted: Dict[str, int] = {}
        self.totalAvgRuntime = 0.0
        self.totalJobsCompleted = 0

        self.targetTime: float = config.targetTime
        if self.targetTime <= 0:
            raise RuntimeError('targetTime (%s) must be a positive integer!' % self.targetTime)
        self.betaInertia = config.betaInertia
        if not 0.0 <= self.betaInertia <= 0.9:
            raise RuntimeError('betaInertia (%f) must be between 0.0 and 0.9!' % self.betaInertia)


        # Pull scaling information from the provisioner.
        self.nodeShapeToType = provisioner.getAutoscaledInstanceShapes()
        self.instance_types = list(self.nodeShapeToType.values())
        self.nodeShapes = list(self.nodeShapeToType.keys())

        self.ignoredNodes: Set[str] = set()

        # A *deficit* exists when we have more jobs that can run on preemptible
        # nodes than we have preemptible nodes. In order to not block these jobs,
        # we want to increase the number of non-preemptible nodes that we have and
        # need for just non-preemptible jobs. However, we may still
        # prefer waiting for preemptible instances to come available.
        # To accommodate this, we set the delta to the difference between the number
        # of provisioned preemptible nodes and the number of nodes that were requested.
        # Then, when provisioning non-preemptible nodes of the same type, we attempt to
        # make up the deficit.
        self.preemptibleNodeDeficit = {instance_type: 0 for instance_type in self.instance_types}

        # Keeps track of the last raw (i.e. float, not limited by
        # max/min nodes) estimates of the number of nodes needed for
        # each node shape. NB: we start with an estimate of 0, so
        # scaling up is smoothed as well.
        self.previousWeightedEstimate = {nodeShape: 0.0 for nodeShape in self.nodeShapes}

        assert len(self.nodeShapes) > 0

        # Minimum/maximum number of either preemptible or non-preemptible nodes in the cluster
        minNodes = config.minNodes
        if minNodes is None:
            minNodes = [0 for node in self.instance_types]
        maxNodes = config.maxNodes
        while len(maxNodes) < len(self.instance_types):
            # Pad out the max node counts if we didn't get one per type.
            maxNodes.append(maxNodes[0])
        while len(minNodes) < len(self.instance_types):
            # Pad out the min node counts with 0s, so we can have fewer than
            # the node types without crashing.
            minNodes.append(0)
        self.minNodes = dict(zip(self.nodeShapes, minNodes))
        self.maxNodes = dict(zip(self.nodeShapes, maxNodes))

        self.nodeShapes.sort()

        # Nodes might not actually provide all the resources of their nominal shapes
        self.node_shapes_after_overhead = self.nodeShapes if config.assume_zero_overhead else [self._reserve_overhead(s) for s in self.nodeShapes]
        self.without_overhead = {k: v for k, v in zip(self.node_shapes_after_overhead, self.nodeShapes)}

        #Node shape to number of currently provisioned nodes
        totalNodes: Dict[Shape, int] = defaultdict(int)
        if isinstance(leader.batchSystem, AbstractScalableBatchSystem) and leader.provisioner:
            for preemptible in (True, False):
                nodes: List["Node"] = []
                for nodeShape, instance_type in self.nodeShapeToType.items():
                    nodes_thisType = leader.provisioner.getProvisionedWorkers(instance_type=instance_type,
                                                                              preemptible=preemptible)
                    totalNodes[nodeShape] += len(nodes_thisType)
                    nodes.extend(nodes_thisType)

                self.setStaticNodes(nodes, preemptible)

        logger.debug('Starting with the following nodes in the cluster: %s' % totalNodes)

        if not sum(config.maxNodes) > 0:
            raise RuntimeError('Not configured to create nodes of any type.')

    def _round(self, number: float) -> int:
        """
        Helper function for rounding-as-taught-in-school (X.5 rounds to X+1 if positive).
        Python 3 now rounds 0.5 to whichever side is even (i.e. 2.5 rounds to 2).

        :param int number: a float to round.
        :return: closest integer to number, rounding ties away from 0.
        """

        sign = 1 if number >= 0 else -1

        rounded = int(round(number))
        nextRounded = int(round(number + 1 * sign))

        if nextRounded == rounded:
            # We rounded X.5 to even, and it was also away from 0.
            return rounded
        elif nextRounded == rounded + 1 * sign:
            # We rounded normally (we are in Python 2)
            return rounded
        elif nextRounded == rounded + 2 * sign:
            # We rounded X.5 to even, but it was towards 0.
            # Go away from 0 instead.
            return rounded + 1 * sign
        else:
            # If we get here, something has gone wrong.
            raise RuntimeError(f"Could not round {number}")

    def _reserve_overhead(self, full_node: Shape) -> Shape:
        """
        Given the shape of an entire virtual machine, return the Shape of the
        resources we expect to actually be available to the backing scheduler
        (i.e. Mesos or Kubernetes) to run jobs on. Some disk may need to be
        used to install the OS, and some memory may be needed to run the
        kernel, Mesos, a Kubelet, Kubernetes system pods, and so on.

        Because shapes are used as dictionary keys, this function must be 1 to
        1: any two distinct inputs (that are actual node shapes and not
        pathological inputs off by a single byte) must produce distinct
        outputs.
        """

        # See
        # https://cloud.google.com/kubernetes-engine/docs/concepts/cluster-architecture#memory_cpu
        # for how EKS works. We are going to just implement their scheme and
        # hope it is a good enough overestimate of actual overhead.
        smaller = copy.copy(full_node)

        # Take away some memory to account for how much the scheduler is going to use.
        smaller.memory -= self._gke_overhead(smaller.memory)
        # And some disk to account for the OS and possibly superuser reservation
        # TODO: Figure out if the disk is an OS disk of a scratch disk
        smaller.disk -= self._disk_overhead(smaller.disk)

        logger.debug('Node shape %s can hold jobs of shape %s', full_node, smaller)

        return smaller

    def _gke_overhead(self, memory_bytes: int) -> int:
        """
        Get the number of bytes of memory that Google Kubernetes Engine would
        reserve or lose to the eviction threshold on a machine with the given
        number of bytes of memory, according to
        <https://cloud.google.com/kubernetes-engine/docs/concepts/cluster-architecture#memory_cpu>.

        We assume that this is a good estimate of how much memory will be lost
        to overhead in a Toil cluster, even though we aren't actually running a
        GKE cluster.
        """

        if memory_bytes < RESERVE_SMALL_LIMIT:
            # On machines with just a little memory, use a flat amount of it.
            return RESERVE_SMALL_AMOUNT

        # How many bytes are reserved so far?
        reserved = 0.0
        # How many bytes of memory have we accounted for so far?
        accounted: Union[float, int] = 0
        for breakpoint, fraction in zip(RESERVE_BREAKPOINTS, RESERVE_FRACTIONS):
            # Below each breakpoint, reserve the matching portion of the memory
            # since the previous breakpoint, like a progressive income tax.
            limit = min(breakpoint, memory_bytes)
            reservation = fraction * (limit - accounted)
            logger.debug('Reserve %s of memory between %s and %s', bytes2human(reservation), bytes2human(accounted), bytes2human(limit))
            reserved += reservation
            accounted = limit
            if accounted >= memory_bytes:
                break
        logger.debug('Reserved %s/%s memory for overhead', bytes2human(reserved), bytes2human(memory_bytes))

        return int(reserved) + EVICTION_THRESHOLD

    def _disk_overhead(self, disk_bytes: int) -> int:
        """
        Get the number of bytes of disk that are probably not usable on a
        machine with a disk of the given size.
        """

        # How much do we think we need for OS and possible super-user
        # reservation?
        disk_needed = OS_SIZE + int(0.05 * disk_bytes)

        if disk_bytes <= disk_needed:
            # We don't think we can actually use any of this disk
            logger.warning('All %sB of disk on a node type are likely to be needed by the OS! The node probably cannot do any useful work!', bytes2human(disk_bytes))
            return disk_bytes

        if disk_needed * 2 > disk_bytes:
            logger.warning('A node type has only %sB disk, of which more than half are expected to be used by the OS. Consider using a larger --nodeStorage', bytes2human(disk_bytes))

        return disk_needed


    def getAverageRuntime(self, jobName: str, service: bool = False) -> float:
        if service:
            # We short-circuit service jobs and assume that they will
            # take a very long time, because if they are assumed to
            # take a short time, we may try to pack multiple services
            # into the same core/memory/disk "reservation", one after
            # the other. That could easily lead to underprovisioning
            # and a deadlock, because often multiple services need to
            # be running at once for any actual work to get done.
            return self.targetTime * 24 + 3600
        if jobName in self.jobNameToAvgRuntime:
            #Have seen jobs of this type before, so estimate
            #the runtime based on average of previous jobs of this type
            return self.jobNameToAvgRuntime[jobName]
        elif self.totalAvgRuntime > 0:
            #Haven't seen this job yet, so estimate its runtime as
            #the average runtime of all completed jobs
            return self.totalAvgRuntime
        else:
            #Have no information whatsoever
            return 1.0

    def addCompletedJob(self, job: JobDescription, wallTime: int) -> None:
        """
        Adds the shape of a completed job to the queue, allowing the scalar to use the last N
        completed jobs in factoring how many nodes are required in the cluster.
        :param toil.job.JobDescription job: The description of the completed job
        :param int wallTime: The wall-time taken to complete the job in seconds.
        """

        #Adjust average runtimes to include this job.
        if job.jobName in self.jobNameToAvgRuntime:
            prevAvg = self.jobNameToAvgRuntime[job.jobName]
            prevNum = self.jobNameToNumCompleted[job.jobName]
            self.jobNameToAvgRuntime[job.jobName] = float(prevAvg*prevNum + wallTime)/(prevNum + 1)
            self.jobNameToNumCompleted[job.jobName] += 1
        else:
            self.jobNameToAvgRuntime[job.jobName] = wallTime
            self.jobNameToNumCompleted[job.jobName] = 1

        self.totalJobsCompleted += 1
        self.totalAvgRuntime = float(self.totalAvgRuntime * (self.totalJobsCompleted - 1) + \
                                     wallTime)/self.totalJobsCompleted

    def setStaticNodes(self, nodes: List["Node"], preemptible: bool) -> None:
        """
        Used to track statically provisioned nodes. This method must be called
        before any auto-scaled nodes are provisioned.

        These nodes are treated differently than auto-scaled nodes in that they should
        not be automatically terminated.

        :param nodes: list of Node objects
        """
        prefix = 'non-' if not preemptible else ''
        logger.debug("Adding %s to %spreemptible static nodes", nodes, prefix)
        if nodes is not None:
            self.static[preemptible] = {node.privateIP : node for node in nodes}

    def getStaticNodes(self, preemptible: bool) -> Dict[str, "Node"]:
        """
        Returns nodes set in setStaticNodes().

        :param preemptible:
        :return: Statically provisioned nodes.
        """
        return self.static[preemptible]

    def smoothEstimate(self, nodeShape: Shape, estimatedNodeCount: int) -> int:
        """
        Smooth out fluctuations in the estimate for this node compared to previous runs.

        Returns an integer.
        """
        weightedEstimate = (1 - self.betaInertia) * estimatedNodeCount + \
                           self.betaInertia * self.previousWeightedEstimate[nodeShape]
        self.previousWeightedEstimate[nodeShape] = weightedEstimate
        return self._round(weightedEstimate)

    def getEstimatedNodeCounts(
        self, queuedJobShapes: List[Shape], currentNodeCounts: Dict[Shape, int]
    ) -> Tuple[Dict[Shape, int], Dict[Shape, List[FailedConstraint]]]:
        """
        Given the resource requirements of queued jobs and the current size of the cluster.

        Returns a dict mapping from nodeShape to the number of nodes we want in
        the cluster right now, and a dict from job shapes that are too big to run
        on any node to reasons why.
        """

        # Do the bin packing with overhead reserved
        nodesToRunQueuedJobs, could_not_fit = binPacking(
            jobShapes=queuedJobShapes,
            nodeShapes=self.node_shapes_after_overhead,
            goalTime=self.targetTime
        )
        
        # Then translate back to get results in terms of full nodes without overhead.
        nodesToRunQueuedJobs = {self.without_overhead[k]: v for k, v in nodesToRunQueuedJobs.items()}

        estimatedNodeCounts = {}
        for nodeShape in self.nodeShapes:
            instance_type = self.nodeShapeToType[nodeShape]

            logger.debug(f"Nodes of type {instance_type} to run queued jobs: {nodesToRunQueuedJobs[nodeShape]}")
            # Actual calculation of the estimated number of nodes required
            estimatedNodeCount = 0 if nodesToRunQueuedJobs[nodeShape] == 0 \
                else max(1, self._round(nodesToRunQueuedJobs[nodeShape]))
            logger.debug("Estimating %i nodes of shape %s" % (estimatedNodeCount, nodeShape))

            # Use inertia parameter to smooth out fluctuations according to an exponentially
            # weighted moving average.
            estimatedNodeCount = self.smoothEstimate(nodeShape, estimatedNodeCount)

            # If we're scaling a non-preemptible node type, we need to see if we have a
            # deficit of preemptible nodes of this type that we should compensate for.
            if not nodeShape.preemptible:
                compensation = self.config.preemptibleCompensation
                assert 0.0 <= compensation <= 1.0
                # The number of nodes we provision as compensation for missing preemptible
                # nodes is the product of the deficit (the number of preemptible nodes we did
                # _not_ allocate) and configuration preference.
                compensationNodes = self._round(self.preemptibleNodeDeficit[instance_type] * compensation)
                if compensationNodes > 0:
                    logger.debug('Adding %d non-preemptible nodes of type %s to compensate for a '
                                'deficit of %d preemptible ones.', compensationNodes,
                                instance_type,
                                self.preemptibleNodeDeficit[instance_type])
                estimatedNodeCount += compensationNodes

            # Tell everyone how big the cluster is
            logger.debug("Currently %i nodes of type %s in cluster" % (currentNodeCounts[nodeShape],
                                                                      instance_type))
            self.leader.toilState.bus.publish(ClusterSizeMessage(instance_type, currentNodeCounts[nodeShape]))
            self.leader.toilState.bus.publish(ClusterDesiredSizeMessage(instance_type, estimatedNodeCount))

            # Bound number using the max and min node parameters
            if estimatedNodeCount > self.maxNodes[nodeShape]:
                logger.debug('Limiting the estimated number of necessary %s (%s) to the '
                             'configured maximum (%s).', instance_type,
                             estimatedNodeCount,
                             self.maxNodes[nodeShape])
                estimatedNodeCount = self.maxNodes[nodeShape]
            elif estimatedNodeCount < self.minNodes[nodeShape]:
                logger.debug('Raising the estimated number of necessary %s (%s) to the '
                            'configured minimum (%s).', instance_type,
                            estimatedNodeCount,
                            self.minNodes[nodeShape])
                estimatedNodeCount = self.minNodes[nodeShape]
            estimatedNodeCounts[nodeShape] = estimatedNodeCount
        return estimatedNodeCounts, could_not_fit

    def updateClusterSize(self, estimatedNodeCounts: Dict[Shape, int]) -> Dict[Shape, int]:
        """
        Given the desired and current size of the cluster, attempts to launch/remove instances to get to the desired size.

        Also attempts to remove ignored nodes that were marked for graceful removal.

        Returns the new size of the cluster.
        """
        newNodeCounts = defaultdict(int)
        for nodeShape, estimatedNodeCount in estimatedNodeCounts.items():
            instance_type = self.nodeShapeToType[nodeShape]

            newNodeCount = self.setNodeCount(instance_type, estimatedNodeCount, preemptible=nodeShape.preemptible)
            # If we were scaling up a preemptible node type and failed to meet
            # our target, we will attempt to compensate for the deficit while scaling
            # non-preemptible nodes of this type.
            if nodeShape.preemptible:
                if newNodeCount < estimatedNodeCount:
                    deficit = estimatedNodeCount - newNodeCount
                    logger.debug('Preemptible scaler detected deficit of %d nodes of type %s.' % (deficit, instance_type))
                    self.preemptibleNodeDeficit[instance_type] = deficit
                else:
                    self.preemptibleNodeDeficit[instance_type] = 0
            newNodeCounts[nodeShape] = newNodeCount

        #Attempt to terminate any nodes that we previously designated for
        #termination, but which still had workers running.
        self._terminateIgnoredNodes()
        return newNodeCounts

    def setNodeCount(
        self,
        instance_type: str,
        numNodes: int,
        preemptible: bool = False,
        force: bool = False,
    ) -> int:
        """
        Attempt to grow or shrink the number of preemptible or non-preemptible worker nodes in
        the cluster to the given value, or as close a value as possible, and, after performing
        the necessary additions or removals of worker nodes, return the resulting number of
        preemptible or non-preemptible nodes currently in the cluster.

        :param instance_type: The instance type to add or remove.

        :param numNodes: Desired size of the cluster

        :param preemptible: whether the added nodes will be preemptible, i.e. whether they
               may be removed spontaneously by the underlying platform at any time.

        :param force: If False, the provisioner is allowed to deviate from the given number
               of nodes. For example, when downsizing a cluster, a provisioner might leave nodes
               running if they have active jobs running on them.

        :return: the number of worker nodes in the cluster after making the necessary
                adjustments. This value should be, but is not guaranteed to be, close or equal to
                the `numNodes` argument. It represents the closest possible approximation of the
                actual cluster size at the time this method returns.
        """
        if not isinstance(self.leader.batchSystem, AbstractScalableBatchSystem):
            raise RuntimeError('Non-scalable batch system abusing a scalable-only function.')
        for attempt in old_retry(predicate=self.provisioner.retryPredicate):
            with attempt:
                nodes = self.getNodes(preemptible)
                logger.debug("Cluster contains %i instances" % len(nodes))

                nodes = {node: nodes[node] for node in nodes if node.nodeType == instance_type}
                ignoredNodes = [node for node in nodes if node.privateIP in self.ignoredNodes]
                numIgnoredNodes = len(ignoredNodes)
                numCurrentNodes = len(nodes)
                logger.debug("Cluster contains %i instances of type %s (%i ignored and draining jobs until "
                             "they can be safely terminated)" % (numCurrentNodes, instance_type, numIgnoredNodes))
                if not force:
                    delta = numNodes - (numCurrentNodes - numIgnoredNodes)
                else:
                    delta = numNodes - numCurrentNodes
                if delta > 0 and numIgnoredNodes > 0:
                    # We can un-ignore a few nodes to compensate for the additional nodes we want.
                    numNodesToUnignore = min(delta, numIgnoredNodes)
                    logger.debug('Unignoring %i nodes because we want to scale back up again.' % numNodesToUnignore)
                    delta -= numNodesToUnignore

                    for node in ignoredNodes[:numNodesToUnignore]:
                        self.ignoredNodes.remove(node.privateIP)
                        self.leader.batchSystem.unignoreNode(node.privateIP)
                if delta > 0:
                    logger.info('Adding %i %s nodes to get to desired cluster size of %i.',
                                delta,
                                'preemptible' if preemptible else 'non-preemptible',
                                numNodes)
                    numNodes = numCurrentNodes + self._addNodes(instance_type, numNodes=delta,
                                                                preemptible=preemptible)
                elif delta < 0:
                    logger.info('Removing %i %s nodes to get to desired cluster size of %i.', -delta, 'preemptible' if preemptible else 'non-preemptible', numNodes)
                    numNodes = numCurrentNodes - self._removeNodes(nodes,
                                                                   instance_type=instance_type,
                                                                   num_nodes=-delta,
                                                                   preemptible=preemptible,
                                                                   force=force)
                elif force:
                    logger.debug('Cluster already at desired size of %i. Nothing to do.', numNodes)
                else:
                    logger.debug('Cluster (minus ignored nodes) already at desired size of %i. Nothing to do.', numNodes)
        return numNodes

    def _addNodes(self, instance_type: str, numNodes: int, preemptible: bool) -> int:
        return self.provisioner.addNodes(nodeTypes={instance_type}, numNodes=numNodes, preemptible=preemptible)

    def _removeNodes(
        self,
        nodes: Dict["Node", NodeInfo],
        instance_type: str,
        num_nodes: int,
        preemptible: bool = False,
        force: bool = False,
    ) -> int:
        # If the batch system is scalable, we can use the number of currently running workers on
        # each node as the primary criterion to select which nodes to terminate.
        if isinstance(self.leader.batchSystem, AbstractScalableBatchSystem):
            # Unless forced, exclude nodes with running workers. Note that it is possible for
            # the batch system to report stale nodes for which the corresponding instance was
            # terminated already. There can also be instances that the batch system doesn't have
            # nodes for yet. We'll ignore those, too, unless forced.
            # TODO: This ignores/overrides the input arg, do we really want to do that???
            nodes = self.getNodes(preemptible)
            # Filter down to nodes of the correct node type

            nodes = {node: nodes[node] for node in nodes if
                     node.nodeType == instance_type}

            filtered_nodes = self.filter_out_static_nodes(nodes, preemptible)
            filtered_nodes = filtered_nodes[:num_nodes]

            # Join nodes and instances on private IP address.
            logger.debug('Nodes considered to terminate: %s', ' '.join(map(str, nodes)))

            # Tell the batch system to stop sending jobs to these nodes
            for (node, nodeInfo) in filtered_nodes:
                self.ignoredNodes.add(node.privateIP)
                self.leader.batchSystem.ignoreNode(node.privateIP)

            if not force:
                # Filter out nodes with jobs still running. These
                # will be terminated in _removeIgnoredNodes later on
                # once all jobs have finished, but they will be ignored by
                # the batch system and cluster scaler from now on
                filtered_nodes = [(node, nodeInfo) for (node, nodeInfo) in filtered_nodes if
                                  nodeInfo and nodeInfo.workers < 1]
            nodes_to_terminate = [node for (node, nodeInfo) in filtered_nodes]
            for node in nodes_to_terminate:
                if node.privateIP in self.ignoredNodes:
                    # TODO: Why are we undoing what was just done above???
                    self.leader.batchSystem.unignoreNode(node.privateIP)
        else:
            # Without load info all we can do is sort instances by time left in billing cycle.
            nodes_to_terminate = sorted(nodes.keys(), key=lambda x: x.remainingBillingInterval())
            nodes_to_terminate = nodes_to_terminate[:num_nodes]
        number_terminated = len(nodes_to_terminate)
        logger.debug('Terminating %i instance(s).', number_terminated)
        for node in nodes_to_terminate:
            if node.privateIP in self.ignoredNodes:
                # TODO: Why are we undoing what was just done above???
                self.ignoredNodes.remove(node.privateIP)
        self.provisioner.terminateNodes(nodes_to_terminate)
        return number_terminated

    def _terminateIgnoredNodes(self) -> None:
        """
        Terminate straggling nodes designated for termination,
        but which still have workers running.
        """
        if not isinstance(self.leader.batchSystem, AbstractScalableBatchSystem):
            raise RuntimeError('Non-scalable batch system abusing a scalable-only function.')

        # start with a dictionary of all nodes and filter down
        nodes = self.getNodes()

        # remove any nodes that have already been terminated from ignored nodes
        all_node_ips = [node.privateIP for node in nodes]
        terminated_node_ips = {ip for ip in self.ignoredNodes if ip not in all_node_ips}

        # unignore terminated nodes
        for ip in terminated_node_ips:
            self.ignoredNodes.remove(ip)
            self.leader.batchSystem.unignoreNode(ip)

        logger.debug("There are %i nodes being ignored by the batch system, "
                     "checking if they can be terminated" % len(self.ignoredNodes))
        nodes = {node: info for node, info in nodes.items() if node.privateIP in self.ignoredNodes}
        nodes = {node: info for node, info in nodes.items() if info and info.workers < 1}
        nodes_to_terminate = list(nodes.keys())

        for node in nodes_to_terminate:
            self.ignoredNodes.remove(node.privateIP)
            self.leader.batchSystem.unignoreNode(node.privateIP)
        self.provisioner.terminateNodes(nodes_to_terminate)

    def filter_out_static_nodes(
            self,
            nodes: Dict["Node", NodeInfo],
            preemptible: bool = False) -> List[Tuple["Node", NodeInfo]]:
        filtered_nodes = []
        for node, nodeInfo in nodes.items():
            if node:
                non = 'non-' if not preemptible else ''
                if node.privateIP in self.getStaticNodes(preemptible):
                    # we don't want to automatically terminate any statically provisioned nodes
                    logger.debug(f'Found {node.privateIP} in {non}preemptible static nodes')
                else:
                    logger.debug(f'Did not find {node.privateIP} in {non}preemptible static nodes')
                    filtered_nodes.append((node, nodeInfo))
        # Sort nodes by number of workers and time left in billing cycle
        filtered_nodes.sort(key=lambda node_nodeInfo: (
            node_nodeInfo[1].workers if node_nodeInfo[1] else 1, node_nodeInfo[0].remainingBillingInterval()))
        return filtered_nodes

    def getNodes(self, preemptible: Optional[bool] = None) -> Dict["Node", NodeInfo]:
        """
        Returns a dictionary mapping node identifiers of preemptible or non-preemptible nodes to
        NodeInfo objects, one for each node.

        This method is the definitive source on nodes in cluster, & is responsible for consolidating
        cluster state between the provisioner & batch system.

        :param bool preemptible: If True (False) only (non-)preemptible nodes will be returned.
               If None, all nodes will be returned.
        """
        if not isinstance(self.leader.batchSystem, AbstractScalableBatchSystem):
            raise RuntimeError('Non-scalable batch system abusing a scalable-only function.')
        # nodes seen within the last 600 seconds (10 minutes)
        recent_nodes = self.leader.batchSystem.getNodes(preemptible, timeout=600)
        # all available nodes
        all_nodes = self.leader.batchSystem.getNodes(preemptible)
        # nodes that are supposedly doing something
        provisioned_nodes = self.provisioner.getProvisionedWorkers(preemptible=preemptible)

        if len(recent_nodes) != len(provisioned_nodes):
            logger.debug("Consolidating state between mesos and provisioner")

        nodeToInfo: Dict["Node", NodeInfo] = {}
        # fixme: what happens if awsFilterImpairedNodes is used?
        # if this assertion is false it means that user-managed nodes are being
        # used that are outside the provisioner's control
        # this would violate many basic assumptions in autoscaling so it currently not allowed
        for node, ip in ((node, node.privateIP) for node in provisioned_nodes):
            if ip not in recent_nodes:
                logger.debug("Worker node at %s is not reporting executor information", ip)

                # get up-to-date information about the node, if available
                info = all_nodes.get(ip)

                # info was found, but the node is not in use
                if not self.leader.batchSystem.nodeInUse(ip) and info:
                    # Node hasn't reported in the last 10 minutes & last we knew there weren't
                    # any tasks running.  We will fake executorInfo with no worker to reflect
                    # this, since otherwise this node will never be considered for termination.
                    info.workers = 0

                # info was not found about the node
                if not info:
                    # Never seen by mesos - 1 of 3 possibilities:
                    # 1) Node is still launching mesos & will come online soon.
                    # 2) No jobs have been assigned to this worker. This means the executor
                    #    was never launched, so we don't even get an executorInfo back indicating
                    #    0 workers running.
                    # 3) Mesos crashed before launching, worker will never come online.
                    #
                    # In all 3 situations it's safe to fake executor info with 0 workers,
                    # since in all cases there are no workers running.
                    info = NodeInfo(coresTotal=1, coresUsed=0, requestedCores=0,
                                    memoryTotal=1, memoryUsed=0, requestedMemory=0,
                                    workers=0)
            else:
                # mesos knows about the ip & we have up-to-date information - easy!
                info = recent_nodes[ip]
            # add info to dict to return
            nodeToInfo[node] = info
        return nodeToInfo

    def shutDown(self) -> None:
        logger.debug('Forcing provisioner to reduce cluster size to zero.')
        for nodeShape in self.nodeShapes:
            preemptible = nodeShape.preemptible
            instance_type = self.nodeShapeToType[nodeShape]
            self.setNodeCount(instance_type=instance_type, numNodes=0, preemptible=preemptible, force=True)

class JobTooBigError(Exception):
    """
    Raised in the scaler thread when a job cannot fit in any available node
    type and is likely to lock up the workflow.
    """
    
    def __init__(self, job: Optional[JobDescription] = None, shape: Optional[Shape] = None, constraints: Optional[List[FailedConstraint]] = None):
        """
        Make a JobTooBigError.
        
        Can have a job, the job's shape, and the limiting resources and amounts. All are optional.
        """
        self.job = job
        self.shape = shape
        self.constraints = constraints if constraints is not None else []
        
        parts = [
            f"The job {self.job}" if self.job else "A job",
            f" with shape {self.shape}" if self.shape else "",
            " is too big for any available node type."
        ]
        
        if self.constraints:
            parts.append(" It could have fit if it only needed ")
            parts.append(", ".join([f"{limit} {resource}" for resource, limit in self.constraints]))
            parts.append(".") 
        
        self.msg = ''.join(parts)
        super().__init__()

    def __str__(self) -> str:
        """
        Stringify the exception, including the message.
        """
        return self.msg

class ScalerThread(ExceptionalThread):
    """
    A thread that automatically scales the number of either preemptible or non-preemptible worker
    nodes according to the resource requirements of the queued jobs.

    The scaling calculation is essentially as follows: start with 0 estimated worker nodes. For
    each queued job, check if we expect it can be scheduled into a worker node before a certain time
    (currently one hour). Otherwise, attempt to add a single new node of the smallest type that
    can fit that job.

    At each scaling decision point a comparison between the current, C, and newly estimated
    number of nodes is made. If the absolute difference is less than beta * C then no change
    is made, else the size of the cluster is adapted. The beta factor is an inertia parameter
    that prevents continual fluctuations in the number of nodes.
    """
    def __init__(self, provisioner: AbstractProvisioner, leader: "Leader", config: Config, stop_on_exception: bool = False) -> None:
        super().__init__(name='scaler')
        self.scaler = ClusterScaler(provisioner, leader, config)
        
        # Indicates that the scaling thread should shutdown
        self.stop = False
        # Indicates that we should stop the thread if we encounter an error.
        # Mostly for testing.
        self.stop_on_exception = stop_on_exception

        self.stats = None
        if config.clusterStats:
            logger.debug("Starting up cluster statistics...")
            self.stats = ClusterStats(leader.config.clusterStats,
                                      leader.batchSystem,
                                      provisioner.clusterName)
            for preemptible in [True, False]:
                self.stats.startStats(preemptible=preemptible)
            logger.debug("...Cluster stats started.")
            
    def check(self) -> None:
        """
        Attempt to join any existing scaler threads that may have died or finished.

        This insures any exceptions raised in the threads are propagated in a timely fashion.
        """
        try:
            self.join(timeout=0)
        except Exception as e:
            logger.exception(e)
            raise

    def shutdown(self) -> None:
        """Shutdown the cluster."""
        self.stop = True
        if self.stats:
            self.stats.shutDownStats()
        self.join()

    def addCompletedJob(self, job: JobDescription, wallTime: int) -> None:
        self.scaler.addCompletedJob(job, wallTime)

    def tryRun(self) -> None:
        if self.scaler.leader.provisioner is None:
            raise RuntimeError('No provisioner found for a scaling cluster '
                               '(cannot access "getProvisionedWorkers").')
        while not self.stop:
            with throttle(self.scaler.config.scaleInterval):
                try:
                    queuedJobs = self.scaler.leader.getJobs()
                    queuedJobShapes = [
                        Shape(wallTime=self.scaler.getAverageRuntime(
                            jobName=job.jobName,
                            service=isinstance(job, ServiceJobDescription)),
                            memory=job.memory,
                            cores=job.cores,
                            disk=job.disk,
                            preemptible=job.preemptible) for job in queuedJobs]
                    currentNodeCounts = {}
                    for nodeShape in self.scaler.nodeShapes:
                        instance_type = self.scaler.nodeShapeToType[nodeShape]
                        currentNodeCounts[nodeShape] = len(
                            self.scaler.leader.provisioner.getProvisionedWorkers(
                                instance_type=instance_type,
                                preemptible=nodeShape.preemptible,
                            )
                        )
                    estimatedNodeCounts, could_not_fit = self.scaler.getEstimatedNodeCounts(
                        queuedJobShapes, currentNodeCounts
                    )
                    self.scaler.updateClusterSize(estimatedNodeCounts)
                    if self.stats:
                        self.stats.checkStats()
                    
                    if len(could_not_fit) != 0: 
                        # If we have any jobs left over that we couldn't fit, complain.
                        bad_job: Optional[JobDescription] = None
                        bad_shape: Optional[Shape] = None
                        for job, shape in zip(queuedJobs, queuedJobShapes):
                            # Try and find an example job with an offending shape
                            if shape in could_not_fit:
                                bad_job = job
                                bad_shape = shape
                                break
                        if bad_shape is None:
                            # If we can't find an offending job, grab an arbitrary offending shape.
                            bad_shape = next(iter(could_not_fit))
                        
                        raise JobTooBigError(job=bad_job, shape=bad_shape, constraints=could_not_fit[bad_shape])
                        
                except:
                    if self.stop_on_exception:
                        logger.critical("Stopping ScalerThread due to an error.")
                        raise
                    else:
                        logger.exception("Exception encountered in scaler thread. Making a best-effort "
                                         "attempt to keep going, but things may go wrong from now on.")
        self.scaler.shutDown()

class ClusterStats:
    def __init__(
        self, path: str, batchSystem: AbstractBatchSystem, clusterName: Optional[str]
    ) -> None:
        logger.debug("Initializing cluster statistics")
        self.stats: Dict[str, Dict[str, List[Dict[str, Any]]]] = {}
        self.statsThreads: List[ExceptionalThread] = []
        self.statsPath = path
        self.stop = False
        self.clusterName = clusterName
        self.batchSystem = batchSystem
        self.scaleable = isinstance(self.batchSystem, AbstractScalableBatchSystem) \
            if batchSystem else False

    def shutDownStats(self) -> None:
        if self.stop:
            return

        def getFileName() -> str:
            extension = '.json'
            file = '%s-stats' % self.clusterName
            counter = 0
            while True:
                suffix = str(counter).zfill(3) + extension
                fullName = os.path.join(self.statsPath, file + suffix)
                if not os.path.exists(fullName):
                    return fullName
                counter += 1
        if self.statsPath and self.scaleable:
            self.stop = True
            for thread in self.statsThreads:
                thread.join()
            fileName = getFileName()
            with open(fileName, 'w') as f:
                json.dump(self.stats, f)

    def startStats(self, preemptible: bool) -> None:
        thread = ExceptionalThread(target=self._gatherStats, args=[preemptible])
        thread.start()
        self.statsThreads.append(thread)

    def checkStats(self) -> None:
        for thread in self.statsThreads:
            # propagate any errors raised in the threads execution
            thread.join(timeout=0)

    def _gatherStats(self, preemptible: bool) -> None:
        def toDict(nodeInfo: NodeInfo) -> Dict[str, Any]:
            # convert NodeInfo object to dict to improve JSON output
            return dict(memory=nodeInfo.memoryUsed,
                        cores=nodeInfo.coresUsed,
                        memoryTotal=nodeInfo.memoryTotal,
                        coresTotal=nodeInfo.coresTotal,
                        requestedCores=nodeInfo.requestedCores,
                        requestedMemory=nodeInfo.requestedMemory,
                        workers=nodeInfo.workers,
                        time=time.time()  # add time stamp
                        )
        if self.scaleable:
            logger.debug("Starting to gather statistics")
            stats: Dict[str, List[Dict[str, Any]]] = {}
            if not isinstance(self.batchSystem, AbstractScalableBatchSystem):
                raise RuntimeError('Non-scalable batch system abusing a scalable-only function.')
            try:
                while not self.stop:
                    nodeInfo = self.batchSystem.getNodes(preemptible)
                    for nodeIP in list(nodeInfo.keys()):
                        nodeStats = nodeInfo[nodeIP]
                        if nodeStats is not None:
                            nodeStatsDict = toDict(nodeStats)
                            try:
                                # if the node is already registered update the
                                # dictionary with the newly reported stats
                                stats[nodeIP].append(nodeStatsDict)
                            except KeyError:
                                # create a new entry for the node
                                stats[nodeIP] = [nodeStatsDict]
                    time.sleep(60)
            finally:
                threadName = 'Preemptible' if preemptible else 'Non-preemptible'
                logger.debug('%s provisioner stats thread shut down successfully.', threadName)
                self.stats[threadName] = stats



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import logging
import os
import shutil
import time
from abc import ABC, abstractmethod
from argparse import ArgumentParser, _ArgumentGroup
from contextlib import contextmanager
from threading import Condition
from typing import (Any,
                    ContextManager,
                    Dict,
                    Iterator,
                    List,
                    NamedTuple,
                    Optional,
                    Set,
                    Union,
                    cast)

from toil.batchSystems.options import OptionSetter
from toil.bus import MessageBus, MessageOutbox
from toil.common import Config, Toil, cacheDirName
from toil.deferred import DeferredFunctionManager
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.job import JobDescription, ParsedRequirement, Requirer
from toil.lib.memoize import memoize
from toil.resource import Resource

logger = logging.getLogger(__name__)

# Value to use as exitStatus in UpdatedBatchJobInfo.exitStatus when status is not available.
EXIT_STATUS_UNAVAILABLE_VALUE = 255

class BatchJobExitReason(enum.IntEnum):
    FINISHED: int = 1
    """Successfully finished."""
    FAILED: int = 2
    """Job finished, but failed."""
    LOST: int = 3
    """Preemptable failure (job's executing host went away)."""
    KILLED: int = 4
    """Job killed before finishing."""
    ERROR: int = 5
    """Internal error."""
    MEMLIMIT: int = 6
    """Job hit batch system imposed memory limit."""

    @classmethod
    def to_string(cls, value: int) -> str:
        """
        Convert to human-readable string.

        Given an int that may be or may be equal to a value from the enum,
        produce the string value of its matching enum entry, or a stringified
        int.
        """
        try:
            return cls(value).name
        except ValueError:
            return str(value)

class UpdatedBatchJobInfo(NamedTuple):
    jobID: int
    exitStatus: int
    """
    The exit status (integer value) of the job. 0 implies successful.

    EXIT_STATUS_UNAVAILABLE_VALUE is used when the exit status is not available
    (e.g. job is lost, or otherwise died but actual exit code was not reported).
    """

    exitReason: Optional[BatchJobExitReason]
    wallTime: Union[float, int, None]

# Information required for worker cleanup on shutdown of the batch system.
class WorkerCleanupInfo(NamedTuple):
    work_dir: Optional[str]
    """Work directory path (where the cache would go) if specified by user"""

    coordination_dir: Optional[str]
    """Coordination directory path (where lock files would go) if specified by user"""

    workflow_id: str
    """Used to identify files specific to this workflow"""

    clean_work_dir: str
    """
    When to clean up the work and coordination directories for a job ('always',
    'onSuccess', 'onError', 'never')
    """

class AbstractBatchSystem(ABC):
    """An abstract base class to represent the interface the batch system must provide to Toil."""
    @classmethod
    @abstractmethod
    def supportsAutoDeployment(cls) -> bool:
        """
        Whether this batch system supports auto-deployment of the user script itself.

        If it does, the :meth:`setUserScript` can be invoked to set the resource
        object representing the user script.

        Note to implementors: If your implementation returns True here, it should also override
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def supportsWorkerCleanup(cls) -> bool:
        """
        Whether this batch system supports worker cleanup.

        Indicates whether this batch system invokes
        :meth:`BatchSystemSupport.workerCleanup` after the last job for a
        particular workflow invocation finishes. Note that the term *worker*
        refers to an entire node, not just a worker process. A worker process
        may run more than one job sequentially, and more than one concurrent
        worker process may exist on a worker node, for the same workflow. The
        batch system is said to *shut down* after the last worker process
        terminates.
        """
        raise NotImplementedError()

    def setUserScript(self, userScript: Resource) -> None:
        """
        Set the user script for this workflow.

        This method must be called before the first job is
        issued to this batch system, and only if :meth:`.supportsAutoDeployment` returns True,
        otherwise it will raise an exception.

        :param userScript: the resource object representing the user script
               or module and the modules it depends on.
        """
        raise NotImplementedError()

    def set_message_bus(self, message_bus: MessageBus) -> None:
        """
        Give the batch system an opportunity to connect directly to the message
        bus, so that it can send informational messages about the jobs it is
        running to other Toil components.
        """

    @abstractmethod
    def issueBatchJob(self, jobDesc: JobDescription, job_environment: Optional[Dict[str, str]] = None) -> int:
        """
        Issues a job with the specified command to the batch system and returns
        a unique jobID.

        :param jobDesc: a toil.job.JobDescription
        :param job_environment: a collection of job-specific environment
                                variables to be set on the worker.

        :return: a unique jobID that can be used to reference the newly issued
                 job
        """
        raise NotImplementedError()

    @abstractmethod
    def killBatchJobs(self, jobIDs: List[int]) -> None:
        """
        Kills the given job IDs. After returning, the killed jobs will not
        appear in the results of getRunningBatchJobIDs. The killed job will not
        be returned from getUpdatedBatchJob.

        :param jobIDs: list of IDs of jobs to kill
        """
        raise NotImplementedError()

    # FIXME: Return value should be a set (then also fix the tests)

    @abstractmethod
    def getIssuedBatchJobIDs(self) -> List[int]:
        """
        Gets all currently issued jobs

        :return: A list of jobs (as jobIDs) currently issued (may be running, or may be
                 waiting to be run). Despite the result being a list, the ordering should not
                 be depended upon.
        """
        raise NotImplementedError()

    @abstractmethod
    def getRunningBatchJobIDs(self) -> Dict[int, float]:
        """
        Gets a map of jobs as jobIDs that are currently running (not just waiting)
        and how long they have been running, in seconds.

        :return: dictionary with currently running jobID keys and how many seconds they have
                 been running as the value
        """
        raise NotImplementedError()

    @abstractmethod
    def getUpdatedBatchJob(self, maxWait: int) -> Optional[UpdatedBatchJobInfo]:
        """
        Returns information about job that has updated its status (i.e. ceased
        running, either successfully or with an error). Each such job will be
        returned exactly once.

        Does not return info for jobs killed by killBatchJobs, although they
        may cause None to be returned earlier than maxWait.

        :param maxWait: the number of seconds to block, waiting for a result

        :return: If a result is available, returns UpdatedBatchJobInfo.
                 Otherwise it returns None. wallTime is the number of seconds (a strictly
                 positive float) in wall-clock time the job ran for, or None if this
                 batch system does not support tracking wall time.
        """
        raise NotImplementedError()

    def getSchedulingStatusMessage(self) -> Optional[str]:
        """
        Get a log message fragment for the user about anything that might be
        going wrong in the batch system, if available.

        If no useful message is available, return None.

        This can be used to report what resource is the limiting factor when
        scheduling jobs, for example. If the leader thinks the workflow is
        stuck, the message can be displayed to the user to help them diagnose
        why it might be stuck.

        :return: User-directed message about scheduling state.
        """

        # Default implementation returns None.
        # Override to provide scheduling status information.
        return None

    @abstractmethod
    def shutdown(self) -> None:
        """
        Called at the completion of a toil invocation.
        Should cleanly terminate all worker threads.
        """
        raise NotImplementedError()

    def setEnv(self, name: str, value: Optional[str] = None) -> None:
        """
        Set an environment variable for the worker process before it is launched.

        The worker process will typically inherit the environment of the machine
        it is running on but this method makes it possible to override specific
        variables in that inherited environment before the worker is launched.
        Note that this mechanism is different to the one used by the worker
        internally to set up the environment of a job. A call to this method affects
        all jobs issued after this method returns. Note to implementors: This
        means that you would typically need to copy the variables before enqueuing a job.

        If no value is provided it will be looked up from the current environment.
        """
        raise NotImplementedError()

    @classmethod
    def add_options(cls, parser: Union[ArgumentParser, _ArgumentGroup]) -> None:
        """
        If this batch system provides any command line options, add them to the given parser.
        """

    @classmethod
    def setOptions(cls, setOption: OptionSetter) -> None:
        """
        Process command line or configuration options relevant to this batch system.

        :param setOption: A function with signature
            setOption(option_name, parsing_function=None, check_function=None, default=None, env=None)
            returning nothing, used to update run configuration as a side effect.
        """

    def getWorkerContexts(self) -> List[ContextManager[Any]]:
        """
        Get a list of picklable context manager objects to wrap worker work in,
        in order.

        Can be used to ask the Toil worker to do things in-process (such as
        configuring environment variables, hot-deploying user scripts, or
        cleaning up a node) that would otherwise require a wrapping "executor"
        process.
        """
        return []


class BatchSystemSupport(AbstractBatchSystem):
    """Partial implementation of AbstractBatchSystem, support methods."""

    def __init__(self, config: Config, maxCores: float, maxMemory: int, maxDisk: int) -> None:
        """
        Initialize initial state of the object.

        :param toil.common.Config config: object is setup by the toilSetup script and
          has configuration parameters for the jobtree. You can add code
          to that script to get parameters for your batch system.

        :param float maxCores: the maximum number of cores the batch system can
          request for any one job

        :param int maxMemory: the maximum amount of memory the batch system can
          request for any one job, in bytes

        :param int maxDisk: the maximum amount of disk space the batch system can
          request for any one job, in bytes
        """
        super().__init__()
        self.config = config
        self.maxCores = maxCores
        self.maxMemory = maxMemory
        self.maxDisk = maxDisk
        self.environment: Dict[str, str] = {}
        if config.workflowID is None:
            raise Exception("config.workflowID must be set")
        else:
            self.workerCleanupInfo = WorkerCleanupInfo(
                work_dir=config.workDir,
                coordination_dir=config.coordination_dir,
                workflow_id=config.workflowID,
                clean_work_dir=config.cleanWorkDir,
            )
        self._outbox: Optional[MessageOutbox] = None

    def check_resource_request(self, requirer: Requirer) -> None:
        """
        Check resource request is not greater than that available or allowed.

        :param requirer: Object whose requirements are being checked

        :param str job_name: Name of the job being checked, for generating a useful error report.

        :param str detail: Batch-system-specific message to include in the error.

        :raise InsufficientSystemResources: raised when a resource is requested in an amount
               greater than allowed
        """
        try:
            for resource, requested, available in [('cores', requirer.cores, self.maxCores),
                                                   ('memory', requirer.memory, self.maxMemory),
                                                   ('disk', requirer.disk, self.maxDisk)]:
                assert requested is not None
                if requested > available:
                    raise InsufficientSystemResources(requirer, resource, available)
            # Handle accelerators in another method that can be overridden separately
            self._check_accelerator_request(requirer)
        except InsufficientSystemResources as e:
            # Add more annotation info to the error
            e.batch_system = self.__class__.__name__ or None
            e.source = self.config.workDir if e.resource == 'disk' else None
            raise e

    def _check_accelerator_request(self, requirer: Requirer) -> None:
        """
        Raise an InsufficientSystemResources error if the batch system can't
        provide the accelerators that are required.

        If a batch system *can* provide accelerators, it should override this
        to say so.
        """
        if len(requirer.accelerators) > 0:
            # By default we assume we can't fulfill any of these
            raise InsufficientSystemResources(requirer, 'accelerators', [], details=[
                'The batch system does not support any accelerators.'
            ])

    def setEnv(self, name: str, value: Optional[str] = None) -> None:
        """
        Set an environment variable for the worker process before it is launched. The worker
        process will typically inherit the environment of the machine it is running on but this
        method makes it possible to override specific variables in that inherited environment
        before the worker is launched. Note that this mechanism is different to the one used by
        the worker internally to set up the environment of a job. A call to this method affects
        all jobs issued after this method returns. Note to implementors: This means that you
        would typically need to copy the variables before enqueuing a job.

        If no value is provided it will be looked up from the current environment.

        :param name: the environment variable to be set on the worker.

        :param value: if given, the environment variable given by name will be set to this value.
            If None, the variable's current value will be used as the value on the worker

        :raise RuntimeError: if value is None and the name cannot be found in the environment
        """
        if value is None:
            try:
                value = os.environ[name]
            except KeyError:
                raise RuntimeError(f"{name} does not exist in current environment")
        self.environment[name] = value

    def set_message_bus(self, message_bus: MessageBus) -> None:
        """
        Give the batch system an opportunity to connect directly to the message
        bus, so that it can send informational messages about the jobs it is
        running to other Toil components.
        """
        # We do in fact send messages to the message bus.
        self._outbox = message_bus.outbox()

    @memoize
    def get_batch_logs_dir(self) -> str:
        """
        Get the directory where the backing batch system should save its logs.

        Only really makes sense if the backing batch system actually saves logs
        to a filesystem; Kubernetes for example does not. Ought to be a
        directory shared between the leader and the workers, if the backing
        batch system writes logs onto the worker's view of the filesystem, like
        many HPC schedulers do.
        """
        if self.config.batch_logs_dir:
            # Use what is specified
            if not os.path.isdir(self.config.batch_logs_dir):
                # But if it doesn't exist, make it exist
                os.makedirs(self.config.batch_logs_dir, exist_ok=True)
            return self.config.batch_logs_dir
        # And if nothing is specified use the workDir.
        return Toil.getToilWorkDir(self.config.workDir)

    def format_std_out_err_path(self, toil_job_id: int, cluster_job_id: str, std: str) -> str:
        """
        Format path for batch system standard output/error and other files
        generated by the batch system itself.

        Files will be written to the batch logs directory (--batchLogsDir,
        defaulting to the Toil work directory) with names containing both the
        Toil and batch system job IDs, for ease of debugging job failures.

        :param: int toil_job_id : The unique id that Toil gives a job.
        :param: cluster_job_id : What the cluster, for example, GridEngine, uses as its internal job id.
        :param: string std : The provenance of the stream (for example: 'err' for 'stderr' or 'out' for 'stdout')

        :rtype: string : Formatted filename; however if self.config.noStdOutErr is true,
             returns '/dev/null' or equivalent.
        """
        if self.config.noStdOutErr:
            return os.devnull

        file_name: str = f'toil_{self.config.workflowID}.{toil_job_id}.{cluster_job_id}.{std}.log'
        logs_dir: str = self.get_batch_logs_dir()
        return os.path.join(logs_dir, file_name)

    def format_std_out_err_glob(self, toil_job_id: int) -> str:
        """
        Get a glob string that will match all file paths generated by format_std_out_err_path for a job.
        """
        file_glob: str = f'toil_{self.config.workflowID}.{toil_job_id}.*.log'
        logs_dir: str = self.get_batch_logs_dir()
        return os.path.join(logs_dir, file_glob)

    @staticmethod
    def workerCleanup(info: WorkerCleanupInfo) -> None:
        """
        Cleans up the worker node on batch system shutdown.

        Also see :meth:`supportsWorkerCleanup`.

        :param WorkerCleanupInfo info: A named tuple consisting of all the relevant information
               for cleaning up the worker.
        """
        logger.debug('Attempting worker cleanup')
        assert isinstance(info, WorkerCleanupInfo)
        assert info.workflow_id is not None
        workflowDir = Toil.getLocalWorkflowDir(info.workflow_id, info.work_dir)
        coordination_dir = Toil.get_local_workflow_coordination_dir(info.workflow_id, info.work_dir, info.coordination_dir)
        DeferredFunctionManager.cleanupWorker(coordination_dir)
        workflowDirContents = os.listdir(workflowDir)
        AbstractFileStore.shutdownFileStore(info.workflow_id, info.work_dir, info.coordination_dir)
        if info.clean_work_dir in ('always', 'onSuccess', 'onError'):
            if workflowDirContents in ([], [cacheDirName(info.workflow_id)]):
                logger.debug('Deleting workflow directory %s', workflowDir)
                shutil.rmtree(workflowDir, ignore_errors=True)
            if coordination_dir != workflowDir:
                # No more coordination to do here either.
                logger.debug('Deleting coordination directory %s', coordination_dir)
                shutil.rmtree(coordination_dir, ignore_errors=True)

class NodeInfo:
    """
    The coresUsed attribute  is a floating point value between 0 (all cores idle) and 1 (all cores
    busy), reflecting the CPU load of the node.

    The memoryUsed attribute is a floating point value between 0 (no memory used) and 1 (all memory
    used), reflecting the memory pressure on the node.

    The coresTotal and memoryTotal attributes are the node's resources, not just the used resources

    The requestedCores and requestedMemory attributes are all the resources that Toil Jobs have reserved on the
    node, regardless of whether the resources are actually being used by the Jobs.

    The workers attribute is an integer reflecting the number of workers currently active workers
    on the node.
    """
    def __init__(self, coresUsed: float, memoryUsed: float,
                 coresTotal: float, memoryTotal: int,
                 requestedCores: float, requestedMemory: int,
                 workers: int) -> None:
        self.coresUsed = coresUsed
        self.memoryUsed = memoryUsed

        self.coresTotal = coresTotal
        self.memoryTotal = memoryTotal

        self.requestedCores = requestedCores
        self.requestedMemory = requestedMemory

        self.workers = workers


class AbstractScalableBatchSystem(AbstractBatchSystem):
    """
    A batch system that supports a variable number of worker nodes.

    Used by :class:`toil.provisioners.clusterScaler.ClusterScaler`
    to scale the number of worker nodes in the cluster
    up or down depending on overall load.
    """

    @abstractmethod
    def getNodes(self, preemptible: Optional[bool] = None, timeout: int = 600) -> Dict[str, NodeInfo]:
        """
        Returns a dictionary mapping node identifiers of preemptible or non-preemptible nodes to
        NodeInfo objects, one for each node.

        :param preemptible: If True (False) only (non-)preemptible nodes will be returned.
               If None, all nodes will be returned.
        """
        raise NotImplementedError()

    @abstractmethod
    def nodeInUse(self, nodeIP: str) -> bool:
        """
        Can be used to determine if a worker node is running any tasks. If the node is doesn't
        exist, this function should simply return False.

        :param nodeIP: The worker nodes private IP address

        :return: True if the worker node has been issued any tasks, else False
        """
        raise NotImplementedError()

    @abstractmethod
    def ignoreNode(self, nodeAddress: str) -> None:
        """
        Stop sending jobs to this node. Used in autoscaling
        when the autoscaler is ready to terminate a node, but
        jobs are still running. This allows the node to be terminated
        after the current jobs have finished.

        :param nodeAddress: IP address of node to ignore.
        """
        raise NotImplementedError()

    @abstractmethod
    def unignoreNode(self, nodeAddress: str) -> None:
        """
        Stop ignoring this address, presumably after
        a node with this address has been terminated. This allows for the
        possibility of a new node having the same address as a terminated one.
        """
        raise NotImplementedError()


class InsufficientSystemResources(Exception):
    def __init__(self, requirer: Requirer, resource: str, available: Optional[ParsedRequirement] = None, batch_system: Optional[str] = None, source: Optional[str] = None, details: List[str] = []) -> None:
        """
        Make a new exception about how we couldn't get enough of something.

        :param requirer: What needed the resources. May have a .jobName string.
        :param resource: The kind of resource requested (cores, memory, disk, accelerators).
        :param requested: The amount requested.
        :param available: The amount actually available.
        :param batch_system: The batch system that could not provide the resource.
        :param source: The place where the resource was to be gotten from. For disk, should be a path.
        :param details: Any extra details about the problem that can be attached to the error.
        """

        self.job_name : Optional[str] = str(requirer)
        self.resource = resource
        self.requested = cast(ParsedRequirement, getattr(requirer, resource))
        self.available = available
        self.batch_system = batch_system
        self.source = source
        self.details = details

    def __str__(self) -> str:
        """
        Explain the exception.
        """

        unit = 'bytes of ' if self.resource in ('disk', 'memory') else ''
        purpose = ' for temporary space' if self.resource == 'disk' else ''
        qualifier = ' free on {self.source}' if self.resource == 'disk' and self.source is not None else ''

        msg = []
        if self.job_name is not None:
            msg.append(f'The job {self.job_name} is requesting ')
        else:
            msg.append(f'Requesting ')
        msg.append(f'{self.requested} {unit}{self.resource}')
        msg.append(purpose)
        if self.available is not None:
            msg.append(f', more than the maximum of {self.available} {unit}{self.resource}{qualifier} that {self.batch_system or "this batch system"} was configured with')
            if self.resource in ('cores', 'memory', 'disk'):
                msg.append(f', or enforced by --max{self.resource.capitalize()}')
        else:
            msg.append(', but that is not available')
        msg.append('.')

        if self.resource == 'disk':
            msg.append(' Try setting/changing the toil option "--workDir" or changing the base temporary directory by setting TMPDIR.')

        for detail in self.details:
            msg.append(' ')
            msg.append(detail)

        return ''.join(msg)


class AcquisitionTimeoutException(Exception):
    """To be raised when a resource request times out."""
    def __init__(self, resource: str, requested: Union[int, float, Set[int]], available: Union[int, float, Set[int]]) -> None:
        """
        Creates an instance of this exception that indicates which resource is insufficient for
        current demands, as well as the resources requested and actually available.
        :param str resource: string representing the resource type
        :param requested: the resources requested that resulted in this exception
        :param available: the resources actually available
        """
        self.requested = requested
        self.available = available
        self.resource = resource


class ResourcePool:
    """
    Represents an integral amount of a resource (such as memory bytes).
    Amounts can be acquired immediately or with a timeout, and released.
    Provides a context manager to do something with an amount of resource
    acquired.
    """
    def __init__(self, initial_value: int, resource_type: str, timeout: float = 5) -> None:
        super().__init__()
        # We use this condition to signal everyone whenever some resource is released.
        # We use its associated lock to guard value.
        self.condition = Condition()
        # This records how much resource is available right now.
        self.value = initial_value
        self.resource_type = resource_type
        self.timeout = timeout

    def acquireNow(self, amount: int) -> bool:
        """
        Reserve the given amount of the given resource.
        Returns True if successful and False if this is not possible immediately.
        """

        with self.condition:
            if amount > self.value:
                return False
            self.value -= amount
            self.__validate()
            return True

    def acquire(self, amount: int) -> None:
        """
        Reserve the given amount of the given resource.
        Raises AcquisitionTimeoutException if this is not possible in under
        self.timeout time.
        """
        with self.condition:
            startTime = time.time()
            while amount > self.value:
                if time.time() - startTime >= self.timeout:
                    # This means the thread timed out waiting for the resource.
                    raise AcquisitionTimeoutException(resource=self.resource_type,
                                                      requested=amount, available=self.value)
                # Allow self.timeout seconds to get the resource, else quit
                # through the above if condition. This wait + timeout is the
                # last thing in the loop such that a request that takes longer
                # than self.timeout due to multiple wakes under the threshold
                # are still honored.
                self.condition.wait(timeout=self.timeout)
            self.value -= amount
            self.__validate()

    def release(self, amount: int) -> None:
        with self.condition:
            self.value += amount
            self.__validate()
            self.condition.notify_all()

    def __validate(self) -> None:
        assert 0 <= self.value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return "ResourcePool(%i)" % self.value

    @contextmanager
    def acquisitionOf(self, amount: int) -> Iterator[None]:
        self.acquire(amount)
        try:
            yield
        finally:
            self.release(amount)


class ResourceSet:
    """
    Represents a collection of distinct resources (such as accelerators).
    Subsets can be acquired immediately or with a timeout, and released.
    Provides a context manager to do something with a set of of resources
    acquired.
    """
    def __init__(self, initial_value: Set[int], resource_type: str, timeout: float = 5) -> None:
        super().__init__()
        # We use this condition to signal everyone whenever some resource is released.
        # We use its associated lock to guard value.
        self.condition = Condition()
        # This records what resources are available right now.
        self.value = initial_value
        self.resource_type = resource_type
        self.timeout = timeout

    def acquireNow(self, subset: Set[int]) -> bool:
        """
        Reserve the given amount of the given resource.
        Returns True if successful and False if this is not possible immediately.
        """

        with self.condition:
            if subset > self.value:
                return False
            self.value -= subset
            return True

    def acquire(self, subset: Set[int]) -> None:
        """
        Reserve the given amount of the given resource.
        Raises AcquisitionTimeoutException if this is not possible in under
        self.timeout time.
        """
        with self.condition:
            startTime = time.time()
            while subset > self.value:
                if time.time() - startTime >= self.timeout:
                    # This means the thread timed out waiting for the resource.
                    raise AcquisitionTimeoutException(resource=self.resource_type,
                                                      requested=subset, available=self.value)
                # Allow self.timeout seconds to get the resource, else quit
                # through the above if condition. This wait + timeout is the
                # last thing in the loop such that a request that takes longer
                # than self.timeout due to multiple wakes under the threshold
                # are still honored.
                self.condition.wait(timeout=self.timeout)
            self.value -= subset

    def release(self, subset: Set[int]) -> None:
        with self.condition:
            self.value |= subset
            self.condition.notify_all()

    def get_free_snapshot(self) -> Set[int]:
        """
        Get a snapshot of what items are free right now.
        May be stale as soon as you get it, but you will need some kind of hint
        to try and do an acquire.
        """
        return set(self.value)

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return "ResourceSet(%s)" % self.value

    @contextmanager
    def acquisitionOf(self, subset: Set[int]) -> Iterator[None]:
        self.acquire(subset)
        try:
            yield
        finally:
            self.release(subset)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
import pkgutil
import warnings
from typing import TYPE_CHECKING, Callable, Dict, List, Sequence, Tuple, Type

from toil.lib.compatibility import deprecated
from toil.lib.memoize import memoize

if TYPE_CHECKING:
    from toil.batchSystems.abstractBatchSystem import AbstractBatchSystem

logger = logging.getLogger(__name__)

#####
# Plugin system/API
#####

def add_batch_system_factory(key: str, class_factory: Callable[[], Type['AbstractBatchSystem']]):
    """
    Adds a batch system to the registry for workflow or plugin-supplied batch systems.

    :param class_factory: A function that returns a batch system class (NOT an instance), which implements :class:`toil.batchSystems.abstractBatchSystem.AbstractBatchSystem`.
    """
    _registry_keys.append(key)
    _registry[key] = class_factory

def get_batch_systems() -> Sequence[str]:
    """
    Get the names of all the availsble batch systems.
    """
    _load_all_plugins()

    return _registry_keys

def get_batch_system(key: str) -> Type['AbstractBatchSystem']:
    """
    Get a batch system class by name.

    :raises: KeyError if the key is not the name of a batch system, and
             ImportError if the batch system's class cannot be loaded.
    """

    return _registry[key]()


DEFAULT_BATCH_SYSTEM = 'single_machine'

#####
# Built-in batch systems
#####

def aws_batch_batch_system_factory():
    from toil.batchSystems.awsBatch import AWSBatchBatchSystem
    return AWSBatchBatchSystem

def gridengine_batch_system_factory():
    from toil.batchSystems.gridengine import GridEngineBatchSystem
    return GridEngineBatchSystem


def lsf_batch_system_factory():
    from toil.batchSystems.lsf import LSFBatchSystem
    return LSFBatchSystem


def single_machine_batch_system_factory():
    from toil.batchSystems.singleMachine import SingleMachineBatchSystem
    return SingleMachineBatchSystem


def mesos_batch_system_factory():
    from toil.batchSystems.mesos.batchSystem import MesosBatchSystem
    return MesosBatchSystem


def slurm_batch_system_factory():
    from toil.batchSystems.slurm import SlurmBatchSystem
    return SlurmBatchSystem

def torque_batch_system_factory():
    from toil.batchSystems.torque import TorqueBatchSystem
    return TorqueBatchSystem


def htcondor_batch_system_factory():
    from toil.batchSystems.htcondor import HTCondorBatchSystem
    return HTCondorBatchSystem


def kubernetes_batch_system_factory():
    from toil.batchSystems.kubernetes import KubernetesBatchSystem
    return KubernetesBatchSystem

#####
# Registry implementation
#####

_registry: Dict[str, Callable[[], Type["AbstractBatchSystem"]]] = {
    'aws_batch'      : aws_batch_batch_system_factory,
    'single_machine' : single_machine_batch_system_factory,
    'grid_engine'    : gridengine_batch_system_factory,
    'lsf'            : lsf_batch_system_factory,
    'mesos'          : mesos_batch_system_factory,
    'slurm'          : slurm_batch_system_factory,
    'torque'         : torque_batch_system_factory,
    'htcondor'       : htcondor_batch_system_factory,
    'kubernetes'     : kubernetes_batch_system_factory
}
_registry_keys = list(_registry.keys())

# We will load any packages starting with this prefix and let them call
# add_batch_system_factory()
_PLUGIN_NAME_PREFIX = "toil_batch_system_"

@memoize
def _load_all_plugins() -> None:
    """
    Load all the batch system plugins that are installed.
    """

    for finder, name, is_pkg in pkgutil.iter_modules():
        # For all installed packages
        if name.startswith(_PLUGIN_NAME_PREFIX):
            # If it is a Toil batch system plugin, import it
            importlib.import_module(name)

#####
# Deprecated API
#####

# We used to directly access these constants, but now the Right Way to use this
# module is add_batch_system_factory() to register and get_batch_systems() to
# get the list/get_batch_system() to get a class by name.


def __getattr__(name):
    """
    Implement a fallback attribute getter to handle deprecated constants.

    See <https://stackoverflow.com/a/48242860>.
    """
    if name == "BATCH_SYSTEM_FACTORY_REGISTRY":
        warnings.warn("BATCH_SYSTEM_FACTORY_REGISTRY is deprecated; use get_batch_system() or add_batch_system_factory()", DeprecationWarning)
        return _registry
    elif name == "BATCH_SYSTEMS":
        warnings.warn("BATCH_SYSTEMS is deprecated; use get_batch_systems()", DeprecationWarning)
        return _registry_keys
    else:
        raise AttributeError(f"Module {__name__} ahs no attribute {name}")


@deprecated(new_function_name="add_batch_system_factory")
def addBatchSystemFactory(key: str, batchSystemFactory: Callable[[], Type['AbstractBatchSystem']]):
    """
    Deprecated method to add a batch system.
    """
    return add_batch_system_factory(key, batchSystemFactory)


#####
# Testing utilities
#####

# We need a snapshot save/restore system for testing. We can't just tamper with
# the globals because module-level globals are their own references, so we
# can't touch this module's global name bindings from a client module.

def save_batch_system_plugin_state() -> Tuple[List[str], Dict[str, Callable[[], Type['AbstractBatchSystem']]]]:
    """
    Return a snapshot of the plugin registry that can be restored to remove
    added plugins. Useful for testing the plugin system in-process with other
    tests.
    """

    snapshot = (list(_registry_keys), dict(_registry))
    return snapshot

def restore_batch_system_plugin_state(snapshot: Tuple[List[str], Dict[str, Callable[[], Type['AbstractBatchSystem']]]]):
    """
    Restore the batch system registry state to a snapshot from
    save_batch_system_plugin_state().
    """

    # We need to apply the snapshot without rebinding the names, because that
    # won't affect modules that imported the names.
    wanted_batch_systems, wanted_registry = snapshot
    _registry_keys.clear()
    _registry_keys.extend(wanted_batch_systems)
    _registry.clear()
    _registry.update(wanted_registry)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create a config file with all default Toil options."""
import logging
import os

from configargparse import ArgParser

from toil.common import generate_config
from toil.statsAndLogging import add_logging_options, set_logging_from_options

logger = logging.getLogger(__name__)


def main() -> None:
    parser = ArgParser()

    parser.add_argument("output", default="config.yaml", help="Filepath to write the config file too. Default=%("
                                                              "default)s")
    add_logging_options(parser)
    options = parser.parse_args()
    set_logging_from_options(options)
    logger.debug("Attempting to write a default config file to %s.", os.path.abspath(options.output))
    generate_config(os.path.abspath(options.output))
    logger.info("Successfully wrote a default config file to %s.", os.path.abspath(options.output))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

import boto3
from moto import mock_iam

from toil.lib.aws import iam
from toil.test import ToilTest

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class IAMTest(ToilTest):
    """Check that given permissions and associated functions perform correctly"""

    def test_permissions_iam(self):
        granted_perms = {'*': {'Action': ['ec2:*', 'iam:*', 's3:*', 'sdb:*'], 'NotAction': []}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is True
        granted_perms = {'*': {'Action': [], 'NotAction': ['s3:*']}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is True

    def test_negative_permissions_iam(self):
        granted_perms = {'*': {'Action': ['ec2:*', 's3:*', 'sdb:*'], 'NotAction': []}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is False
        granted_perms = {'*': {'Action': [], 'NotAction': ['iam:*', 'ec2:*']}}
        assert iam.policy_permissions_allow(granted_perms, iam.CLUSTER_LAUNCHING_PERMISSIONS) is False

    def test_wildcard_handling(self):
        assert iam.permission_matches_any("iam:CreateRole", ['iam:Create**']) is True
        assert iam.permission_matches_any("iam:GetUser", ['iam:???????']) is True
        assert iam.permission_matches_any("iam:ListRoleTags", ['iam:*?*Tags']) is True
        assert iam.permission_matches_any("iam:*", ["*"]) is True
        assert iam.permission_matches_any("ec2:*", ['iam:*']) is False

    @mock_iam
    def test_get_policy_permissions(self):
        mock_iam = boto3.client("iam")

        # username that moto pretends we have from client.get_user()
        user_name = "default_user"
        mock_iam.create_user(UserName=user_name)

        group_name = "default_group"
        mock_iam.create_group(
            GroupName=group_name
        )

        mock_iam.add_user_to_group(
            GroupName=group_name,
            UserName=user_name
        )

        policy_response = mock_iam.create_policy(
            PolicyName="test_iam_createrole",
            PolicyDocument=json.dumps({
                "Version": "2012-10-17",  # represents version language
                "Statement": [
                    {"Effect": "Allow", "Action": "iam:CreateRole", "Resource": "*"}
                ]
            })
        )
        # attached user policy
        mock_iam.attach_user_policy(
            UserName=user_name,
            PolicyArn=policy_response["Policy"]["Arn"]
        )

        # inline user policy
        mock_iam.put_user_policy(
            UserName=user_name,
            PolicyName="test_iam_createinstanceprofile",
            PolicyDocument=json.dumps({
                "Version": "2012-10-17",  # represents version language
                "Statement": [
                    {"Effect": "Allow", "Action": "iam:CreateInstanceProfile", "Resource": "*"}
                ]
            })
        )

        # group policies
        policy_response = mock_iam.create_policy(
            PolicyName="test_iam_taginstanceprofile",
            PolicyDocument=json.dumps({
                "Version": "2012-10-17",  # represents version language
                "Statement": [
                    {"Effect": "Allow", "Action": "iam:TagInstanceProfile", "Resource": "*"}
                ]
            })
        )
        # attached group policy
        mock_iam.attach_group_policy(
            GroupName=group_name,
            PolicyArn=policy_response["Policy"]["Arn"]
        )

        # inline group policy
        mock_iam.put_group_policy(
            GroupName=group_name,
            PolicyName="test_iam_deleterole",
            PolicyDocument=json.dumps({
                "Version": "2012-10-17",  # represents version language
                "Statement": [
                    {"Effect": "Allow", "Action": "iam:DeleteRole", "Resource": "*"}
                ]
            })
        )

        actions_collection = iam.get_policy_permissions("us-west-2")

        actions_set = set(actions_collection["*"]["Action"])
        notactions_set = set(actions_collection["*"]["NotAction"])

        expected_actions = {"iam:CreateRole", "iam:CreateInstanceProfile", "iam:TagInstanceProfile", "iam:DeleteRole"}
        assert actions_set == expected_actions
        assert notactions_set == set()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import fcntl
import logging
import os
import tempfile
from collections import defaultdict
from contextlib import contextmanager
from typing import (IO,
                    Any,
                    Callable,
                    ContextManager,
                    DefaultDict,
                    Dict,
                    Generator,
                    Iterator,
                    List,
                    Literal,
                    Optional,
                    Union,
                    cast,
                    overload)

import dill

from toil.common import getFileSystemSize
from toil.fileStores import FileID
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.job import Job, JobDescription
from toil.jobStores.abstractJobStore import AbstractJobStore
from toil.lib.compatibility import deprecated
from toil.lib.io import make_public_dir, robust_rmtree
from toil.lib.retry import ErrorCondition, retry
from toil.lib.threading import get_process_name, process_name_exists

logger: logging.Logger = logging.getLogger(__name__)


class NonCachingFileStore(AbstractFileStore):
    def __init__(
        self,
        jobStore: AbstractJobStore,
        jobDesc: JobDescription,
        file_store_dir: str,
        waitForPreviousCommit: Callable[[], Any],
    ) -> None:
        super().__init__(jobStore, jobDesc, file_store_dir, waitForPreviousCommit)
        # This will be defined in the `open` method.
        self.jobStateFile: Optional[str] = None
        self.localFileMap: DefaultDict[str, List[str]] = defaultdict(list)

        self.check_for_state_corruption()

    @staticmethod
    def check_for_coordination_corruption(coordination_dir: Optional[str]) -> None:
        """
        Make sure the coordination directory hasn't been deleted unexpectedly.

        Slurm has been known to delete XDG_RUNTIME_DIR out from under processes
        it was promised to, so it is possible that in certain misconfigured
        environments the coordination directory and everything in it could go
        away unexpectedly. We are going to regularly make sure that the things
        we think should exist actually exist, and we are going to abort if they
        do not.
        """

        if coordination_dir and not os.path.exists(coordination_dir):
            raise RuntimeError(
                f'The Toil coordination directory at {coordination_dir} '
                f'was removed while the workflow was running! Please provide a '
                f'TOIL_COORDINATION_DIR or --coordinationDir at a location that '
                f'is safe from automated cleanup during the workflow run.'
            )

    def check_for_state_corruption(self) -> None:
        """
        Make sure state tracking information hasn't been deleted unexpectedly.
        """

        NonCachingFileStore.check_for_coordination_corruption(self.coordination_dir)

        if self.jobStateFile and not os.path.exists(self.jobStateFile):
            raise RuntimeError(
                f'The job state file {self.jobStateFile} '
                f'was removed while the workflow was running! Please provide a '
                f'TOIL_COORDINATION_DIR or --coordinationDir at a location that '
                f'is safe from automated cleanup during the workflow run.'
            )

    @contextmanager
    def open(self, job: Job) -> Generator[None, None, None]:
        startingDir = os.getcwd()
        self.localTempDir: str = make_public_dir(in_directory=self.localTempDir)
        self._removeDeadJobs(self.coordination_dir)
        self.jobStateFile = self._createJobStateFile()
        self.check_for_state_corruption()
        freeSpace, diskSize = getFileSystemSize(self.localTempDir)
        if freeSpace <= 0.1 * diskSize:
            logger.warning(f'Starting job {self.jobName} with less than 10%% of disk space remaining.')
        try:
            os.chdir(self.localTempDir)
            with super().open(job):
                yield
        finally:
            os.chdir(startingDir)
            # Finally delete the job from the worker
            self.check_for_state_corruption()
            try:
                os.remove(self.jobStateFile)
            except FileNotFoundError:
                logger.exception('Job state file %s has gone missing unexpectedly; some cleanup for failed jobs may be getting skipped!', self.jobStateFile)
                pass

    def writeGlobalFile(self, localFileName: str, cleanup: bool=False) -> FileID:
        absLocalFileName = self._resolveAbsoluteLocalPath(localFileName)
        creatorID = str(self.jobDesc.jobStoreID)
        fileStoreID = self.jobStore.write_file(absLocalFileName, creatorID, cleanup)
        if absLocalFileName.startswith(self.localTempDir):
            # Only files in the appropriate directory should become local files
            # we can delete with deleteLocalFile
            self.localFileMap[fileStoreID].append(absLocalFileName)
        return FileID.forPath(fileStoreID, absLocalFileName)

    def readGlobalFile(self, fileStoreID: str, userPath: Optional[str] = None, cache: bool=True, mutable: bool=False,
                            symlink: bool=False) -> str:
        if userPath is not None:
            localFilePath = self._resolveAbsoluteLocalPath(userPath)
            if os.path.exists(localFilePath):
                raise RuntimeError(' File %s ' % localFilePath + ' exists. Cannot Overwrite.')
        else:
            localFilePath = self.getLocalTempFileName()

        self.jobStore.read_file(fileStoreID, localFilePath, symlink=symlink)
        self.localFileMap[fileStoreID].append(localFilePath)
        self.logAccess(fileStoreID, localFilePath)
        return localFilePath

    @overload
    def readGlobalFileStream(
        self,
        fileStoreID: str,
        encoding: Literal[None] = None,
        errors: Optional[str] = None,
    ) -> ContextManager[IO[bytes]]:
        ...

    @overload
    def readGlobalFileStream(
        self, fileStoreID: str, encoding: str, errors: Optional[str] = None
    ) -> ContextManager[IO[str]]:
        ...

    # TODO: This seems to hit https://github.com/python/mypy/issues/11373
    # But that is supposedly fixed.

    @contextmanager # type: ignore
    def readGlobalFileStream(self, fileStoreID: str, encoding: Optional[str] = None, errors: Optional[str] = None) -> Iterator[Union[IO[bytes], IO[str]]]:
        with self.jobStore.read_file_stream(fileStoreID, encoding=encoding, errors=errors) as f:
            self.logAccess(fileStoreID)
            yield f

    @deprecated(new_function_name='export_file')
    def exportFile(self, jobStoreFileID: FileID, dstUrl: str) -> None:
        return self.export_file(jobStoreFileID, dstUrl)

    def export_file(self, file_id: FileID, dst_uri: str) -> None:
        self.jobStore.export_file(file_id, dst_uri)

    def deleteLocalFile(self, fileStoreID: str) -> None:
        try:
            localFilePaths = self.localFileMap.pop(fileStoreID)
        except KeyError:
            raise OSError(errno.ENOENT, "Attempting to delete local copies of a file with none")
        else:
            for localFilePath in localFilePaths:
                os.remove(localFilePath)

    def deleteGlobalFile(self, fileStoreID: str) -> None:
        try:
            self.deleteLocalFile(fileStoreID)
        except OSError as e:
            if e.errno == errno.ENOENT:
                # the file does not exist locally, so no local deletion necessary
                pass
            else:
                raise
        self.filesToDelete.add(str(fileStoreID))

    def waitForCommit(self) -> bool:
        # there is no asynchronicity in this file store so no need to block at all
        return True

    def startCommit(self, jobState: bool = False) -> None:
        # Make sure the previous job is committed, if any
        if self.waitForPreviousCommit is not None:
            self.waitForPreviousCommit()

        # We are going to commit synchronously, so no need to clone a snapshot
        # of the job description or mess with its version numbering.

        if not jobState:
            # All our operations that need committing are job state related
            return

        try:
            # Indicate any files that should be seen as deleted once the
            # update of the job description is visible.
            if len(self.jobDesc.filesToDelete) > 0:
                raise RuntimeError("Job is already in the process of being committed!")
            self.jobDesc.filesToDelete = list(self.filesToDelete)
            # Complete the job
            self.jobStore.update_job(self.jobDesc)
            # Delete any remnant files
            list(map(self.jobStore.delete_file, self.filesToDelete))
            # Remove the files to delete list, having successfully removed the files
            if len(self.filesToDelete) > 0:
                self.jobDesc.filesToDelete = []
                # Update, removing emptying files to delete
                self.jobStore.update_job(self.jobDesc)
        except:
            self._terminateEvent.set()
            raise


    def __del__(self) -> None:
        """
        Cleanup function that is run when destroying the class instance.  Nothing to do since there
        are no async write events.
        """

    @classmethod
    def _removeDeadJobs(cls, coordination_dir: str, batchSystemShutdown: bool=False) -> None:
        """
        Look at the state of all jobs registered in the individual job state files, and handle them
        (clean up the disk)

        :param str coordination_dir: The location of the coordination directory on the node.
        :param bool batchSystemShutdown: Is the batch system in the process of shutting down?
        :return:
        """

        cls.check_for_coordination_corruption(coordination_dir)

        for jobState in cls._getAllJobStates(coordination_dir):
            if not process_name_exists(coordination_dir, jobState['jobProcessName']):
                # We need to have a race to pick someone to clean up.

                try:
                    # Open the directory
                    dirFD = os.open(jobState['jobDir'], os.O_RDONLY)
                except FileNotFoundError:
                    # The cleanup has happened and we can't contest for it
                    continue

                try:
                    # Try and lock it
                    fcntl.lockf(dirFD, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except OSError as e:
                    # We lost the race. Someone else is alive and has it locked.
                    os.close(dirFD)
                else:
                    # We got it
                    logger.warning('Detected that job (%s) prematurely terminated.  Fixing the '
                                   'state of the job on disk.', jobState['jobName'])

                    try:
                        if not batchSystemShutdown:
                            logger.debug("Deleting the stale working directory.")
                            # Delete the old work directory if it still exists.  Do this only during
                            # the life of the program and dont' do it during the batch system
                            # cleanup. Leave that to the batch system cleanup code.
                            robust_rmtree(jobState['jobDir'])
                    finally:
                        fcntl.lockf(dirFD, fcntl.LOCK_UN)
                        os.close(dirFD)

    @classmethod
    def _getAllJobStates(cls, coordination_dir: str) -> Iterator[Dict[str, str]]:
        """
        Generator function that deserializes and yields the job state for every job on the node,
        one at a time.

        :param coordination_dir: The location of the coordination directory on the node.

        :return: dict with keys (jobName,  jobProcessName, jobDir)
        """
        jobStateFiles = []

        cls.check_for_coordination_corruption(coordination_dir)

        # Note that the directory may contain files whose names are not decodable to Unicode.
        # So we need to work in bytes.
        for entry in os.scandir(os.fsencode(coordination_dir)):
            # For each job state file in the coordination directory
            if entry.name.endswith(b'.jobState'):
                # This is the state of a job
                jobStateFiles.append(os.fsdecode(entry.path))

        for fname in jobStateFiles:
            try:
                yield NonCachingFileStore._readJobState(fname)
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # This is a FileNotFoundError.
                    # job finished & deleted its jobState file since the jobState files were discovered
                    continue
                elif e.errno == 5:
                    # This is a OSError: [Errno 5] Input/output error (jobStatefile seems to disappear 
                    # on network file system sometimes)
                    continue
                else:
                    raise

    @staticmethod
    # Retry on any OSError except FileNotFoundError, which we throw immediately
    @retry(errors=[
        OSError,
        ErrorCondition(
            error=FileNotFoundError,
            retry_on_this_condition=False
        )])
    def _readJobState(jobStateFileName: str) -> Dict[str, str]:
        with open(jobStateFileName, 'rb') as fH:
            state = dill.load(fH)
        return cast(Dict[str, str], state)

    def _createJobStateFile(self) -> str:
        """
        Create the job state file for the current job and fill in the required
        values.

        Places the file in the coordination directory.

        :return: Path to the job state file
        :rtype: str
        """
        self.check_for_state_corruption()
        jobState = {'jobProcessName': get_process_name(self.coordination_dir),
                    'jobName': self.jobName,
                    'jobDir': self.localTempDir}
        try:
            (fd, jobStateFile) = tempfile.mkstemp(suffix='.jobState.tmp', dir=self.coordination_dir)
        except Exception as e:
            raise RuntimeError("Could not make state file in " + self.coordination_dir) from e
        with open(fd, 'wb') as fH:
            # Write data
            dill.dump(jobState, fH)
        # Drop suffix
        jobStateFile = jobStateFile[:-len('.tmp')]
        # Put in place
        os.rename(jobStateFile + '.tmp', jobStateFile)
        return jobStateFile

    @classmethod
    def shutdown(cls, shutdown_info: str) -> None:
        """
        :param shutdown_info: The coordination directory.
        """
        cls._removeDeadJobs(shutdown_info, batchSystemShutdown=True)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Debug tool for copying files contained in a toil jobStore."""
import argparse
import logging
import os.path
import sys
from typing import Optional

from toil.common import Config, Toil, parser_with_common_options
from toil.jobStores.fileJobStore import FileJobStore
from toil.lib.resources import glob
from toil.lib.conversions import strtobool
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)


def fetchJobStoreFiles(jobStore: FileJobStore, options: argparse.Namespace) -> None:
    """
    Takes a list of file names as glob patterns, searches for these within a
    given directory, and attempts to take all of the files found and copy them
    into options.localFilePath.

    :param jobStore: A fileJobStore object.
    :param options.fetch: List of file glob patterns to search
        for in the jobStore and copy into options.localFilePath.
    :param options.localFilePath: Local directory to copy files into.
    :param options.jobStore: The path to the jobStore directory.
    """

    # TODO: Implement the necessary methods in the job store class and stop
    # globbing around inside it. Does this even work?

    for jobStoreFile in options.fetch:
        jobStoreHits = glob(directoryname=options.jobStore,
                            glob_pattern=jobStoreFile)
        for jobStoreFileID in jobStoreHits:
            logger.debug(f"Copying job store file: {jobStoreFileID} to {options.localFilePath[0]}")
            jobStore.read_file(jobStoreFileID,
                               os.path.join(options.localFilePath[0],
                                           os.path.basename(jobStoreFileID)),
                               symlink=options.useSymlinks)


def printContentsOfJobStore(job_store: FileJobStore, job_id: Optional[str] = None) -> None:
    """
    Fetch a list of all files contained in the job store if nameOfJob is not
    declared, otherwise it only prints out the names of files for that specific
    job for which it can find a match.  Also creates a log file of these file
    names in the current directory.

    :param job_store: Job store to ask for files from.
    :param job_id: Default is None, which prints out all files in the jobStore.
        If specified, it will print all jobStore files that have been written
        to the jobStore by that job.
    """

    # TODO: Implement the necessary methods for job stores other than
    # FileJobStore.

    if job_id:
        logFile = job_id.replace("/", "_") + "_fileset.txt"
    else:
        logFile = "jobstore_files.txt"

    list_of_files = job_store.list_all_file_names(for_job=job_id)
    if os.path.exists(logFile):
        os.remove(logFile)
    for gfile in sorted(list_of_files):
        if job_id:
            logger.debug(f"{job_id} File: {os.path.basename(gfile)}")
        else:
            logger.debug(f"File: {os.path.basename(gfile)}")
        with open(logFile, "a+") as f:
            f.write(os.path.basename(gfile))
            f.write("\n")


def main() -> None:
    parser = parser_with_common_options(jobstore_option=True, prog="toil debug-file")
    parser.add_argument("--localFilePath",
                        nargs=1,
                        help="Location to which to copy job store files.")
    parser.add_argument("--fetch",
                        nargs="+",
                        help="List of job-store files to be copied locally."
                             "Use either explicit names (i.e. 'data.txt'), or "
                             "specify glob patterns (i.e. '*.txt')")
    parser.add_argument("--listFilesInJobStore", type=strtobool,
                        help="Prints a list of the current files in the jobStore.")
    parser.add_argument("--fetchEntireJobStore", type=strtobool,
                        help="Copy all job store files into a local directory.")
    parser.add_argument("--useSymlinks", type=strtobool,
                        help="Creates symlink 'shortcuts' of files in the localFilePath"
                             " instead of hardlinking or copying, where possible.  If this is"
                             " not possible, it will copy the files (shutil.copyfile()).")

    # Load the jobStore
    options = parser.parse_args()
    set_logging_from_options(options)
    config = Config()
    config.setOptions(options)
    jobStore = Toil.resumeJobStore(config.jobStore)
    logger.debug("Connected to job store: %s", config.jobStore)

    if options.fetch:
        # Copy only the listed files locally

        if isinstance(jobStore, FileJobStore):
            logger.debug("Fetching local files: %s", options.fetch)
            fetchJobStoreFiles(jobStore=jobStore, options=options)
        else:
            # The user asked for something we can't do yet.
            # Tell them no but don't stack trace.
            logger.critical("Can only fetch by name or glob from file-based job stores")
            sys.exit(1)

    elif options.fetchEntireJobStore:
        # Copy all jobStore files locally

        if isinstance(jobStore, FileJobStore):
            logger.debug("Fetching all local files.")
            options.fetch = "*"
            fetchJobStoreFiles(jobStore=jobStore, options=options)
        else:
            logger.critical("Can only fetch by name or glob from file-based job stores")
            sys.exit(1)

    if options.listFilesInJobStore:
        # Log filenames and create a file containing these names in cwd

        if isinstance(jobStore, FileJobStore):
            printContentsOfJobStore(job_store=jobStore)
        else:
            logger.critical("Can only list files from file-based job stores")
            sys.exit(1)

    # TODO: We can't actually do *anything* for non-file job stores.


if __name__ == "__main__":
    main()



import os
import uuid

from toil.provisioners import cluster_factory
from toil.test.provisioners.clusterTest import AbstractClusterTest


class CactusIntegrationTest(AbstractClusterTest):
    """
    Run the Cactus Integration test on a Kubernetes AWS cluster
    """

    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        self.clusterName = "cactus-test-" + str(uuid.uuid4())
        self.leaderNodeType = "t2.medium"
        self.clusterType = "kubernetes"

    def setUp(self):
        super().setUp()
        self.jobStore = f"aws:{self.awsRegion()}:cluster-{uuid.uuid4()}"

    def test_cactus_integration(self):
        # Make a cluster with worker nodes
        self.createClusterUtil(args=["--nodeTypes=t2.xlarge", "-w=1-3"])
        # get the leader so we know the IP address - we don't need to wait since create cluster
        # already ensures the leader is running
        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        self.leader = self.cluster.getLeader()

        CACTUS_COMMIT_SHA = os.environ["CACTUS_COMMIT_SHA"] or "f5adf4013326322ae58ef1eccb8409b71d761583"  # default cactus commit

        # command to install and run cactus on the cluster
        cactus_command = ("python -m virtualenv --system-site-packages venv && "
                          ". venv/bin/activate && "
                          "git clone https://github.com/ComparativeGenomicsToolkit/cactus.git --recursive && "
                          "cd cactus && "
                          "git fetch origin && "
                          f"git checkout {CACTUS_COMMIT_SHA} && "
                          "git submodule update --init --recursive && "
                          "pip install --upgrade 'setuptools<66' pip && "
                          "pip install --upgrade . && "
                          "pip install --upgrade numpy psutil && "
                          "time cactus --batchSystem kubernetes --retryCount=3 "
                          f"--consCores 2 --binariesMode singularity --clean always {self.jobStore} "
                          "examples/evolverMammals.txt examples/evolverMammals.hal --root mr --defaultDisk 8G --logDebug")

        # run cactus
        self.sshUtil(
            [
                "bash",
                "-c",
                cactus_command
            ]
        )




# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import pytest

from toil.lib.aws import build_tag_dict_from_env
from toil.test import ToilTest

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

class TagGenerationTest(ToilTest):
    """
    Test for tag generation from environment variables
    """
    def test_build_tag(self):
        environment = dict()
        environment["TOIL_OWNER_TAG"] = "😀"
        environment["TOIL_AWS_TAGS"] = None
        tag_dict = build_tag_dict_from_env(environment)
        assert(tag_dict == {'Owner': '😀'})

    def test_empty_aws_tags(self):
        environment = dict()
        environment["TOIL_OWNER_TAG"] = None
        environment["TOIL_AWS_TAGS"] = "{}"
        tag_dict = build_tag_dict_from_env(environment)
        assert (tag_dict == dict())

    def test_incorrect_json_object(self):
        with pytest.raises(SystemExit):
            environment = dict()
            environment["TOIL_OWNER_TAG"] = None
            environment["TOIL_AWS_TAGS"] = "231"
            tag_dict = build_tag_dict_from_env(environment)

    def test_incorrect_json_emoji(self):
        with pytest.raises(SystemExit):
            environment = dict()
            environment["TOIL_OWNER_TAG"] = None
            environment["TOIL_AWS_TAGS"] = "😀"
            tag_dict = build_tag_dict_from_env(environment)

    def test_build_tag_with_tags(self):
        environment = dict()
        environment["TOIL_OWNER_TAG"] = "😀"
        environment["TOIL_AWS_TAGS"] = '{"1": "2", " ":")"}'
        tag_dict = build_tag_dict_from_env(environment)
        assert(tag_dict == {'Owner': '😀', '1': '2', ' ': ')'})






# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from toil.common import Config
from toil.job import CheckpointJobDescription, JobDescription
from toil.jobStores.fileJobStore import FileJobStore
from toil.test import ToilTest
from toil.worker import nextChainable


class WorkerTests(ToilTest):
    """Test miscellaneous units of the worker."""
    def setUp(self):
        super().setUp()
        path = self._getTestJobStorePath()
        self.jobStore = FileJobStore(path)
        self.config = Config()
        self.config.jobStore = 'file:%s' % path
        self.jobStore.initialize(self.config)
        self.jobNumber = 0

    def testNextChainable(self):
        """Make sure chainable/non-chainable jobs are identified correctly."""
        def createTestJobDesc(memory, cores, disk, preemptible, checkpoint):
            """
            Create a JobDescription with no command (representing a Job that
            has already run) and return the JobDescription.
            """
            name = 'job%d' % self.jobNumber
            self.jobNumber += 1

            descClass = CheckpointJobDescription if checkpoint else JobDescription
            jobDesc = descClass(requirements={'memory': memory, 'cores': cores, 'disk': disk, 'preemptible': preemptible}, jobName=name)

            # Assign an ID
            self.jobStore.assign_job_id(jobDesc)

            # Save and return the JobDescription
            return self.jobStore.create_job(jobDesc)

        for successorType in ['addChild', 'addFollowOn']:
            # Try with the branch point at both child and follow-on stages

            # Identical non-checkpoint jobs should be chainable.
            jobDesc1 = createTestJobDesc(1, 2, 3, True, False)
            jobDesc2 = createTestJobDesc(1, 2, 3, True, False)
            getattr(jobDesc1, successorType)(jobDesc2.jobStoreID)
            chainable = nextChainable(jobDesc1, self.jobStore, self.config)
            self.assertNotEqual(chainable, None)
            self.assertEqual(jobDesc2.jobStoreID, chainable.jobStoreID)

            # Identical checkpoint jobs should not be chainable.
            jobDesc1 = createTestJobDesc(1, 2, 3, True, False)
            jobDesc2 = createTestJobDesc(1, 2, 3, True, True)
            getattr(jobDesc1, successorType)(jobDesc2.jobStoreID)
            self.assertEqual(None, nextChainable(jobDesc1, self.jobStore, self.config))

            # If there is no child we should get nothing to chain.
            jobDesc1 = createTestJobDesc(1, 2, 3, True, False)
            self.assertEqual(None, nextChainable(jobDesc1, self.jobStore, self.config))

            # If there are 2 or more children we should get nothing to chain.
            jobDesc1 = createTestJobDesc(1, 2, 3, True, False)
            jobDesc2 = createTestJobDesc(1, 2, 3, True, False)
            jobDesc3 = createTestJobDesc(1, 2, 3, True, False)
            getattr(jobDesc1, successorType)(jobDesc2.jobStoreID)
            getattr(jobDesc1, successorType)(jobDesc3.jobStoreID)
            self.assertEqual(None, nextChainable(jobDesc1, self.jobStore, self.config))

            # If there is an increase in resource requirements we should get nothing to chain.
            reqs = {'memory': 1, 'cores': 2, 'disk': 3, 'preemptible': True, 'checkpoint': False}
            for increased_attribute in ('memory', 'cores', 'disk'):
                jobDesc1 = createTestJobDesc(**reqs)
                reqs[increased_attribute] += 1
                jobDesc2 = createTestJobDesc(**reqs)
                getattr(jobDesc1, successorType)(jobDesc2.jobStoreID)
                self.assertEqual(None, nextChainable(jobDesc1, self.jobStore, self.config))

            # A change in preemptability from True to False should be disallowed.
            jobDesc1 = createTestJobDesc(1, 2, 3, True, False)
            jobDesc2 = createTestJobDesc(1, 2, 3, False, True)
            getattr(jobDesc1, successorType)(jobDesc2.jobStoreID)
            self.assertEqual(None, nextChainable(jobDesc1, self.jobStore, self.config))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
import tempfile

import pytest

from toil.test import ToilTest

from toil.lib.resources import glob
from toil.test import slow
from toil.version import python

logger = logging.getLogger(__name__)


def workflow_debug_jobstore() -> str:
    job_store_path = os.path.join(tempfile.mkdtemp(), "toilWorkflowRun")
    subprocess.check_call(
        [
            python,
            os.path.abspath("src/toil/test/utils/ABCWorkflowDebug/debugWorkflow.py"),
            job_store_path,
        ]
    )
    return job_store_path


@slow
def testJobStoreContents():
    """
    Test toilDebugFile.printContentsOfJobStore().

    Runs a workflow that imports 'B.txt' and 'mkFile.py' into the
    jobStore.  'A.txt', 'C.txt', 'ABC.txt' are then created.  This checks to
    make sure these contents are found in the jobStore and printed.
    """
    contents = ["A.txt", "B.txt", "C.txt", "ABC.txt", "mkFile.py"]

    subprocess.check_call(
        [
            python,
            os.path.abspath("src/toil/utils/toilDebugFile.py"),
            workflow_debug_jobstore(),
            "--logDebug",
            "--listFilesInJobStore=True",
        ]
    )
    jobstoreFileContents = os.path.abspath("jobstore_files.txt")
    files = []
    match = 0
    with open(jobstoreFileContents) as f:
        for line in f:
            files.append(line.strip())
    for xfile in files:
        for expected_file in contents:
            if xfile.endswith(expected_file):
                match = match + 1
    logger.debug(files)
    logger.debug(contents)
    logger.debug(match)
    # C.txt will match twice (once with 'C.txt', and once with 'ABC.txt')
    assert match == 6
    os.remove(jobstoreFileContents)


def fetchFiles(symLink: bool, jobStoreDir: str, outputDir: str):
    """
    Fn for testFetchJobStoreFiles() and testFetchJobStoreFilesWSymlinks().

    Runs a workflow that imports 'B.txt' and 'mkFile.py' into the
    jobStore.  'A.txt', 'C.txt', 'ABC.txt' are then created.  This test then
    attempts to get a list of these files and copy them over into our
    output diectory from the jobStore, confirm that they are present, and
    then delete them.
    """
    contents = ["A.txt", "B.txt", "C.txt", "ABC.txt", "mkFile.py"]
    cmd = [
        python,
        os.path.abspath("src/toil/utils/toilDebugFile.py"),
        jobStoreDir,
        "--fetch",
        "*A.txt",
        "*B.txt",
        "*C.txt",
        "*ABC.txt",
        "*mkFile.py",
        f"--localFilePath={outputDir}",
        f"--useSymlinks={symLink}",
    ]
    print(cmd)
    subprocess.check_call(cmd)
    for xfile in contents:
        matchingFilesFound = glob(glob_pattern="*" + xfile, directoryname=outputDir)
        assert len(matchingFilesFound) >= 1
        for fileFound in matchingFilesFound:
            assert fileFound.endswith(xfile) and os.path.exists(fileFound)
            if fileFound.endswith("-" + xfile):
                os.remove(fileFound)


# expected run time = 4s
def testFetchJobStoreFiles() -> None:
    """Test toilDebugFile.fetchJobStoreFiles() symlinks."""
    job_store_dir = workflow_debug_jobstore()
    output_dir = os.path.join(os.path.dirname(job_store_dir), "testoutput")
    os.makedirs(output_dir, exist_ok=True)
    for symlink in (True, False):
        fetchFiles(symLink=symlink, jobStoreDir=job_store_dir, outputDir=output_dir)

class DebugJobTest(ToilTest):
    """
    Test the toil debug-job command.
    """

    def _get_job_store_and_job_id(self):
        """
        Get a job store and the ID of a failing job within it.
        """

        # First make a job store.
        job_store = os.path.join(self._createTempDir(), "tree")

        logger.info("Running workflow that always fails")
        try:
            # Run an always-failign workflow
            subprocess.check_call([
                python,
                os.path.abspath("src/toil/test/docs/scripts/example_alwaysfail.py"),
                "--retryCount=0",
                "--logCritical",
                "--disableProgress",
                job_store
            ], stderr=subprocess.DEVNULL)
            raise RuntimeError("Failing workflow succeeded!")
        except subprocess.CalledProcessError:
            # Should fail to run
            logger.info("Task failed successfully")
            pass

        # Get the job ID.
        # TODO: This assumes a lot about the FileJobStore. Use the MessageBus instead?
        job_id = "kind-explode/" + os.listdir(os.path.join(job_store, "jobs/kind-explode"))[0]

        return job_store, job_id

    def test_run_job(self):
        """
        Make sure that we can use toil debug-job to try and run a job in-process.
        """

        job_store, job_id = self._get_job_store_and_job_id()

        logger.info("Trying to rerun job %s", job_id)

        # Rerun the job, which should fail again
        output = subprocess.check_output([
            "toil",
            "debug-job",
            "--logDebug",
            job_store,
            job_id
        ], stderr=subprocess.STDOUT)
        # Even if the job fails, the attempt to run it will succeed.
        log = output.decode('utf-8')
        assert "Boom!" in log, f"Did not find the expected exception message in: {log}"


    def test_print_job_info(self):
        """
        Make sure that we can use --printJobInfo to get information on a job from a job store.
        """

        job_store, job_id = self._get_job_store_and_job_id()

        logger.info("Trying to print job info for job %s", job_id)

        # Print the job info and make sure that doesn't crash.
        subprocess.check_call([
            "toil",
            "debug-job",
            "--logDebug",
            job_store,
            "--printJobInfo",
            job_id
        ])





# Copyright (C) 2020-2021 UCSC Computational Genomics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable


def get_version(iterable: Iterable[str]) -> str:
    """
    Get the version of the WDL document.

    :param iterable: An iterable that contains the lines of a WDL document.
    :return: The WDL version used in the workflow.
    """
    if isinstance(iterable, str):
        iterable = iterable.split('\n')

    for line in iterable:
        line = line.strip()
        # check if the first non-empty, non-comment line is the version statement
        if line and not line.startswith('#'):
            if line.startswith('version '):
                return line[8:].strip()
            break
    # only draft-2 doesn't contain the version declaration
    return 'draft-2'



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import os
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from inspect import getsource
from io import BytesIO
from textwrap import dedent
from unittest.mock import MagicMock, patch
from zipfile import ZipFile

from toil import inVirtualEnv
from toil.resource import ModuleDescriptor, Resource, ResourceException
from toil.test import ToilTest
from toil.version import exactPython


@contextmanager
def tempFileContaining(content, suffix=''):
    """
    Write a file with the given contents, and keep it on disk as long as the context is active.
    :param str content: The contents of the file.
    :param str suffix: The extension to use for the temporary file.
    """
    fd, path = tempfile.mkstemp(suffix=suffix)
    try:
        encoded = content.encode('utf-8')
        assert os.write(fd, encoded) == len(encoded)
    except:
        os.close(fd)
        raise
    else:
        os.close(fd)
        yield path
    finally:
        os.unlink(path)


class ResourceTest(ToilTest):
    """Test module descriptors and resources derived from them."""
    def testStandAlone(self):
        self._testExternal(moduleName='userScript', pyFiles=('userScript.py', 'helper.py'))

    def testPackage(self):
        self._testExternal(moduleName='foo.userScript', pyFiles=('foo/__init__.py',
                                                                 'foo/userScript.py',
                                                                 'foo/bar/__init__.py',
                                                                 'foo/bar/helper.py'))

    def testVirtualEnv(self):
        self._testExternal(moduleName='foo.userScript',
                           virtualenv=True,
                           pyFiles=('foo/__init__.py',
                                    'foo/userScript.py',
                                    'foo/bar/__init__.py',
                                    'foo/bar/helper.py',
                                    'de/pen/dency.py',
                                    'de/__init__.py',
                                    'de/pen/__init__.py'))

    def testStandAloneInPackage(self):
        self.assertRaises(ResourceException,
                          self._testExternal,
                          moduleName='userScript',
                          pyFiles=('__init__.py', 'userScript.py', 'helper.py'))

    def _testExternal(self, moduleName, pyFiles, virtualenv=False):
        dirPath = self._createTempDir()
        if virtualenv:
            self.assertTrue(inVirtualEnv())
            # --never-download prevents silent upgrades to pip, wheel and setuptools
            subprocess.check_call(['virtualenv', '--never-download', '--python', exactPython, dirPath])
            sitePackages = os.path.join(dirPath, 'lib',
                                        exactPython,
                                        'site-packages')
            # tuple assignment is necessary to make this line immediately precede the try:
            oldPrefix, sys.prefix, dirPath = sys.prefix, dirPath, sitePackages
        else:
            oldPrefix = None
        try:
            for relPath in pyFiles:
                path = os.path.join(dirPath, relPath)
                os.makedirs(os.path.dirname(path), exist_ok=True)
                with open(path, 'w') as f:
                    f.write('pass\n')
            sys.path.append(dirPath)
            try:
                userScript = importlib.import_module(moduleName)
                try:
                    self._test(userScript.__name__,
                               expectedContents=pyFiles,
                               allowExtraContents=True)
                finally:
                    del userScript
                    while moduleName:
                        del sys.modules[moduleName]
                        self.assertFalse(moduleName in sys.modules)
                        moduleName = '.'.join(moduleName.split('.')[:-1])

            finally:
                sys.path.remove(dirPath)
        finally:
            if oldPrefix:
                sys.prefix = oldPrefix

    def testBuiltIn(self):
        # Create a ModuleDescriptor for the module containing ModuleDescriptor, i.e. toil.resource
        module_name = ModuleDescriptor.__module__
        self.assertEqual(module_name, 'toil.resource')
        self._test(module_name, shouldBelongToToil=True)

    def _test(self, module_name,
              shouldBelongToToil=False, expectedContents=None, allowExtraContents=True):
        module = ModuleDescriptor.forModule(module_name)
        # Assert basic attributes and properties
        self.assertEqual(module.belongsToToil, shouldBelongToToil)
        self.assertEqual(module.name, module_name)
        if shouldBelongToToil:
            self.assertTrue(module.dirPath.endswith('/src'))

        # Before the module is saved as a resource, localize() and globalize() are identity
        # methods. This should log.warnings.
        self.assertIs(module.localize(), module)
        self.assertIs(module.globalize(), module)
        # Create a mock job store ...
        jobStore = MagicMock()
        # ... to generate a fake URL for the resource ...
        url = 'file://foo.zip'
        jobStore.getSharedPublicUrl.return_value = url
        # ... and save the resource to it.
        resource = module.saveAsResourceTo(jobStore)
        # Ensure that the URL generation method is actually called, ...
        jobStore.getSharedPublicUrl.assert_called_once_with(sharedFileName=resource.pathHash)
        # ... and that ensure that write_shared_file_stream is called.
        jobStore.write_shared_file_stream.assert_called_once_with(shared_file_name=resource.pathHash,
                                                                  encrypted=False)
        # Now it gets a bit complicated: Ensure that the context manager returned by the
        # jobStore's write_shared_file_stream() method is entered and that the file handle yielded
        # by the context manager is written to once with the zipped source tree from which
        # 'toil.resource' was originally imported. Keep the zipped tree around such that we can
        # mock the download later.
        file_handle = jobStore.write_shared_file_stream.return_value.__enter__.return_value
        # The first 0 index selects the first call of write(), the second 0 selects positional
        # instead of keyword arguments, and the third 0 selects the first positional, i.e. the
        # contents. This is a bit brittle since it assumes that all the data is written in a
        # single call to write(). If more calls are made we can easily concatenate them.
        zipFile = file_handle.write.call_args_list[0][0][0]
        self.assertTrue(zipFile.startswith(b'PK'))  # the magic header for ZIP files

        # Check contents if requested
        if expectedContents is not None:
            with ZipFile(BytesIO(zipFile)) as _zipFile:
                actualContents = set(_zipFile.namelist())
                if allowExtraContents:
                    self.assertTrue(actualContents.issuperset(expectedContents))
                else:
                    self.assertEqual(actualContents, expectedContents)

        self.assertEqual(resource.url, url)
        # Now we're on the worker. Prepare the storage for localized resources
        Resource.prepareSystem()
        try:
            # Register the resource for subsequent lookup.
            resource.register()
            # Lookup the resource and ensure that the result is equal to but not the same as the
            # original resource. Lookup will also be used when we localize the module that was
            # originally used to create the resource.
            localResource = Resource.lookup(module._resourcePath)
            self.assertEqual(resource, localResource)
            self.assertIsNot(resource, localResource)
            # Now show that we can localize the module using the registered resource. Set up a mock
            # urlopen() that yields the zipped tree ...
            mock_urlopen = MagicMock()
            mock_urlopen.return_value.read.return_value = zipFile
            with patch('toil.resource.urlopen', mock_urlopen):
                # ... and use it to download and unpack the resource
                localModule = module.localize()
            # The name should be equal between original and localized resource ...
            self.assertEqual(module.name, localModule.name)
            # ... but the directory should be different.
            self.assertNotEqual(module.dirPath, localModule.dirPath)
            # Show that we can 'undo' localization. This is necessary when the user script's jobs
            #  are invoked on the worker where they generate more child jobs.
            self.assertEqual(localModule.globalize(), module)
        finally:
            Resource.cleanSystem()

    def testNonPyStandAlone(self):
        """
        Asserts that Toil enforces the user script to have a .py or .pyc extension because that's
        the only way auto-deployment can re-import the module on a worker. See

        https://github.com/BD2KGenomics/toil/issues/631 and
        https://github.com/BD2KGenomics/toil/issues/858
        """

        def script():
            from configargparse import ArgumentParser

            from toil.common import Toil
            from toil.job import Job

            def fn():
                pass

            if __name__ == '__main__':
                parser = ArgumentParser()
                Job.Runner.addToilOptions(parser)
                options = parser.parse_args()
                job = Job.wrapFn(fn, memory='10M', cores=0.1, disk='10M')
                with Toil(options) as toil:
                    toil.start(job)

        scriptBody = dedent('\n'.join(getsource(script).split('\n')[1:]))
        shebang = '#! %s\n' % sys.executable
        with tempFileContaining(shebang + scriptBody) as scriptPath:
            self.assertFalse(scriptPath.endswith(('.py', '.pyc')))
            os.chmod(scriptPath, 0o755)
            jobStorePath = scriptPath + '.jobStore'
            process = subprocess.Popen([scriptPath, jobStorePath], stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()
            self.assertTrue('The name of a user script/module must end in .py or .pyc.' in stderr.decode('utf-8'))
            self.assertNotEqual(0, process.returncode)
            self.assertFalse(os.path.exists(jobStorePath))



from configargparse import ArgParser

from toil.common import addOptions, Toil
from toil.test import ToilTest


class OptionsTest(ToilTest):
    """
    Class to test functionality of all Toil options
    """
    def test_default_caching_slurm(self):
        """
        Test to ensure that caching will be set to false when running on Slurm
        :return:
        """
        parser = ArgParser()
        addOptions(parser, jobstore_as_flag=True, wdl=False, cwl=False)
        test_args = ["--jobstore=example-jobstore", "--batchSystem=slurm"]
        options = parser.parse_args(test_args)
        with Toil(options) as toil:
            caching_value = toil.config.caching
        self.assertEqual(caching_value, False)

    def test_caching_option_priority(self):
        """
        Test to ensure that the --caching option takes priority over the default_caching() return value
        :return:
        """
        parser = ArgParser()
        addOptions(parser, jobstore_as_flag=True, wdl=False, cwl=False)
        # the kubernetes batchsystem (and I think all batchsystems including singlemachine) return False
        # for default_caching
        test_args = ["--jobstore=example-jobstore", "--batchSystem=kubernetes", "--caching=True"]
        options = parser.parse_args(test_args)
        with Toil(options) as toil:
            caching_value = toil.config.caching
        self.assertEqual(caching_value, True)
